"""
The Upgrade Manager is a bit too heavy-handed and it violates some good design principles.

This tool is meant to perform an update of a single node in a few different ways.
"""
from antelope import enum, MultipleReferences, NoReference  # is this *always* an interactive tool?


class NodeUpgraderException(Exception):
    pass


class NoCandidates(NodeUpgraderException):
    pass


class TooManyCandidates(NodeUpgraderException):
    pass


# need to escape regex-active characters
nonregexp = str.maketrans('()[]\\/*', '.......')


class NodeUpdater:
    _strategy = 'match_name_and_spatial_scope'

    _strategies = ('match_name_and_spatial_scope',
                   'match_name',
                   'same_id',
                   'targets',
                   'targets_match_spatial_scope',
                   'node_targets')

    @property
    def strategy(self):
        return self._strategy

    @strategy.setter
    def strategy(self, value):
        if value is not None:
            if value in self._strategies:
                print('setting strategy "%s"' % value)
                self._strategy = value
            else:
                print('ignoring unrecognized strategy "%s"' % value)

    def __init__(self, branch, query, strategy=None):
        """
        This accepts a single FragmentBranch, which includes a node (FragmentRef) and an anchor (Anchor), with
        a scenario name.  This branch is SUPPOSED to represent an existing termination.
        :param branch:
        :param query: used to obtain upgrade targets
        """
        self._branch = branch
        self._q = query
        self.strategy = strategy
        self._candidates = self._exception = self._rx = self._rxs = None

    @property
    def current(self):
        return self._branch

    """
    different upgrade strategies
    """
    def match_name(self):
        """
        Match the name of the current anchor
        :return:
        """
        name = '^%s$' % self.current.anchor.term_node['name'].translate(nonregexp)
        return self._q.processes(name='^%s$' % name)

    def match_name_and_spatial_scope(self):
        """
        Match the name and spatial scope of the current anchor
        :return:
        """
        name = '^%s$' % self.current.anchor.term_node['name'].translate(nonregexp)
        return self._q.processes(name='^%s$' % name,
                                 spatialscope='^%s$' % self.current.anchor.term_node['spatialscope'])

    def same_id(self):
        """
        Retrieve the process with the same external_ref from a different query
        :return:
        """
        return [self._q.get(self.current.anchor.term_node.external_ref)]

    def targets(self):
        """
        Retrieve targets that match the current anchor (flow and direction)
        :return:
        """
        return self._q.targets(self.current.anchor.term_flow.external_ref, direction=self.current.node.direction)

    def targets_match_spatial_scope(self):
        """
        Retrieve targets that match the current anchor (flow and direction), filter by current anchor's spatial scope
        :return:
        """
        return filter(lambda x: x['spatialscope'] == self.current.anchor.term_node['spatialscope'],
                      self._q.targets(self.current.anchor.term_flow.external_ref, direction=self.current.node.direction)
                      )

    def node_targets(self):
        """
        Retrieve targets that match the current node (flow, disregarding direction)
        :return:
        """
        return self._q.targets(self.current.node.flow)

    def _run_attempt(self, strategy):
        """
        :return:
        """
        if strategy is None:
            strategy = self.strategy
        self._candidates = enum(getattr(self, strategy)())
        if len(self._candidates) == 0:
            raise NoCandidates

    def attempt(self):
        self._run_attempt(self.strategy)

    def rxs(self, n=0):
        return self._candidates[n].references()

    def pick(self, n=0, ref_flow=None):
        """
        subclass: process with ambiguous ref
        :param n:
        :param ref_flow:
        :return:
        """
        self._rxs = tuple(self._candidates[n].references())
        try:
            self._rx = self._candidates[n].reference(ref_flow)  # if ref flow is None and len(rx) is 1, will return it
        except MultipleReferences as m:
            if ref_flow is None:
                try:
                    self._rx = self._candidates[n].reference(self.current.anchor.term_flow)
                    return
                except NoReference:
                    pass
            raise m
        except NoReference:
            raise

    def upgrade(self, n=None, apply_scenario=None):
        """
        Attempts to apply the currently selected
        :param n: which of the current candidates to select
        :param apply_scenario: must be specified.  To run the upgrade in-place (i.e. same scenario as
        observed branch), specify 'in place. To upgrade the default node (i.e. scenario=None), specify 'default'
        :return:
        """
        if n is None:
            if self._rx is None:
                if self._candidates is None:
                    self.attempt()
                if len(self._candidates) > 1:
                    raise TooManyCandidates()
                elif len(self._candidates) == 0:
                    raise NoCandidates()
                else:
                    n = 0
                self.pick(n)
        else:
            self.pick(n)

        if self._branch.scenario is None:
            if apply_scenario is None:
                raise ValueError('Must specify "default" to overwrite default scenario')
            elif apply_scenario == 'default':
                _apply_scenario = None
            else:
                _apply_scenario = apply_scenario
        else:
            if apply_scenario is None:
                raise ValueError('must specify "in place" to overwrite current scenario %s or'
                                 ' "default" to set default scenario' % self._branch.scenario)
            elif apply_scenario == 'in place':
                _apply_scenario = self._branch.scenario
            elif apply_scenario == 'default':
                _apply_scenario = None
            else:
                _apply_scenario = apply_scenario

        print('%s => %s [%s]' % (self._branch.node, self._rx.process, _apply_scenario))
        self._branch.node.clear_termination(_apply_scenario)
        self._branch.node.terminate(self._rx.process, scenario=_apply_scenario,
                                    term_flow=self._rx.flow, descend=self._branch.anchor.descend)



"""
MFA Study

This is a revision of the prototype 'mfa_models' framework for building a nested LCA study with an observed
MFA at its core.

The NestedLcaStudy creates a three-tier system with a study container, a logistics container, and an activity
container, which is terminated to the study driver (nominally an MFA fragment).  The NestedLcaStudy includes
tools to build "routes" that attach to the logistics / study containers to link MFA flows to LCA models.  This is
done with an explicit mapping of flow to route.

In fact, the flow is supposed to plug into a MARKET of routes.
"""
from .lc_mfa_study import NestedLcaStudy
from .dynamic_unit_study import DynamicUnitLcaStudy, DynamicUnitSpec
from .observed_mfa_study import ObservedMfaStudy, StudySpec



"""
Antelope LCIA analysis tools

The purpose of this package is to support LCIA data quality evaluation by comparing LCI + LCIA flow coverage
across methods and data sources.

"""

from .lcia_eval import LciaEval
from .flow_comparator import FlowComparator
# from .screening import process_screen, substance_screen, show_top_n
from .xlsx_lcia import QdbGSheetClient
from .traci import traci_2_combined_eutrophication

from antelope import EntityNotFound
from synonym_dict import MergeError


S3_QDB_SHEET_ID = '1gBi9690IcMRf4B8oAgGQm02yaj459i4vHDmMH2SXlVE'


class Scope3Qdb(QdbGSheetClient):
    def __init__(self, fg, credential_file=None, **kwargs):
        super(Scope3Qdb, self).__init__(fg, S3_QDB_SHEET_ID, credential_file=credential_file, **kwargs)


def add_synonym_sets(cat, list_of_sets):
    """
    An algorithm for adding / merging sets of synonyms
    :param cat:
    :param list_of_sets:
    :return:
    """
    for terms in list_of_sets:
        try:
            cat.lcia_engine.add_terms('flow', *terms)
        except MergeError:
            t = list(terms)
            lost = []
            while 1:
                # this is clearly broken
                try:
                    cat.lcia_engine.get_flowable(t[0])
                except KeyError:
                    try:
                        lost.append(t.pop(0))
                    except IndexError:
                        print('somehow we found no terms to add in %s' % lost)
                        break
                    continue
                break
            if t:
                cat.lcia_engine.merge_flowables(*t)


class AdvancedLcia(object):

    s3_qdb = None

    def __init__(self, cat, lcia_fg, google_credentials=None):
        self._cat = cat
        if hasattr(lcia_fg, 'new_quantity'):
            self._lcia = lcia_fg
        else:
            self._lcia = cat.foreground(lcia_fg, create=True)
        if google_credentials:
            self.s3_qdb_init(google_credentials)

    def s3_qdb_init(self, google_credentials):
        self.s3_qdb = Scope3Qdb(self._lcia, google_credentials)

    def get_s3_qdb_indicator(self, indicator, external_ref=None):
        return self.s3_qdb.update_cfs(indicator, external_ref)

    def traci_combined_eutrophication(self, external_ref='q_eutrophication', omit_n2=True):
        traci = self._cat.query('lcia.traci.2.1')
        qe = traci_2_combined_eutrophication(traci, self._lcia, external_ref=external_ref, omit_n2=omit_n2)
        return qe

    def retrieve_lcia_method(self, origin, external_ref, **kwargs):
        try:
            q = self._lcia.get(external_ref)
        except EntityNotFound:
            q = self._cat.query(origin).get(external_ref)
        for k, v in kwargs.items():
            q[k] = v
        return q



from .charts import ModelGraph
from .observer import QuickAndEasy, MissingValue

from time import time


class DumbTimer(object):
    """
    a centralized timer
    """
    def __init__(self, head='==', tail=None):
        self.t0 = time()
        self.head = head
        self.tail = tail or ''

    @property
    def _t(self):
        return time() - self.t0

    @property
    def _ts(self):
        return '(%5.2f s)' % self._t

    def format(self, string):
        return '%s%s %s%s' % (self.head, self._ts, string, self.tail)

    def check(self, string):
        print(self.format(string))



from .lca_study import LcaStudy, DuplicateRoute
from .model_updater import XlsxForegroundUpdater
from .scenario_updater import XlsxScenarioUpdater
from .ecoinvent_grids import EcoinventGrids, LEVELS


def create_grids(cat, fg, ei_version, *grids, levels=LEVELS, chooser=True, **kwargs):
    """
    This function ensures that locality-based electricity grids are present as fragment models in the
    supplied foreground.

    :param cat: required to access ecoinvent
    :param fg: where to build the grid fragments
    :param ei_version: ecoinvent version
    :param grids: remaining positional params are grid geographies to lookup
    :param levels: voltage levels. Default is ('low', 'medium', 'high')
    :param chooser: [True] whether to create a grid chooser in the foreground
    :param kwargs: passed to EcoinventGrids:
     - 'path' - where to serialize the lookup file
     - 'reset' - whether to blow away an existing lookup file
     - 'local' - if present, prefix to prepend to ecoinvent origins.
    :return:
    """
    egl = EcoinventGrids(cat, version=ei_version, levels=levels, **kwargs)
    for region in grids:
        egl.create_loc_grids(fg, region, rshort=region, chooser=chooser)



"""
OK revisiting this chart for the first time in N years-- it is actually quite nice, but two things:
 1- stacking different scenarios sharing the same quantity with vertically-aligned ordinate axes is not really workable
    because the chart becomes too long.  But it is a clever idea and maybe it should be supported, just not
    without a proper faceted chart design. i.e. this needs a long table, seaborn-style
 2- in fact, what would work is having charts with different quantities with horizontally aligned abcissae would be
    great for contribution analysis
 3- the charts need to show uncertainty/variability per stage- this is absolutely a make-or-break feature

Proposed function signature:

 WaterfallChart(*results, stages=None, ...) without uncertainty
 WaterfallChart(*(res, res_hi, res_lo), stages=None, ...) with uncertainty

The outer layer needs to create a figure, then iterate through the axes
- then a waterfall drawing class that just takes the data and draws it
- stage labeling happens outside

For now: strip out everything related to vertical stacking.

Class for making waterfall charts. There is an inheritance structure to be found in these charts somewhere..


"""

import matplotlib as mpl
import matplotlib.pyplot as plt
import colorsys
from itertools import accumulate
from collections import defaultdict
from math import ceil

# from math import floor

from .base import save_plot, net_color

net_style = {
    'edgecolor': 'none'
}

# mpl.rcParams['patch.force_edgecolor'] = False
mpl.rcParams['errorbar.capsize'] = 3
mpl.rcParams['grid.color'] = 'k'
mpl.rcParams['grid.linestyle'] = ':'
mpl.rcParams['grid.linewidth'] = 0.5
# mpl.rcParams['lines.linewidth'] = 1.0
mpl.rcParams['axes.autolimit_mode'] = 'round_numbers'
mpl.rcParams['axes.xmargin'] = 0
mpl.rcParams['axes.ymargin'] = 0


def _fade_color(color):
    hsv = colorsys.rgb_to_hsv(*color)
    new_hue = (hsv[0] + 0.086) % 1

    return colorsys.hsv_to_rgb(new_hue, hsv[1]*0.65, hsv[2]*0.8)


def random_color(seed, sat=0.65, val=0.95, offset=14669):

    # hue = ((int('%.4s' % uuid, 16) + offset) % 65536) / 65536
    hue = (hash((seed, offset)) % 65536)/65536
    return colorsys.hsv_to_rgb(hue, sat, val)


def grab_stages(*results, sort=None, count=None):
    stages = set()

    def _first_score(_comp):
        for r in results:
            try:
                return r[_comp].cumulative_result
            except KeyError:
                continue
        return 0
    for i in results:
        stages = stages.union(i.keys())
    if sort is None:
        out = sorted(stages, key=_first_score, reverse=True)
    else:
        out = sorted(stages, key=sort)
    if count is None:
        return out
    return out[:count]


def grab_names(*results, field=None, default='Other'):
    d = dict()
    for r in results:
        for k in r.keys():
            if k in d:
                continue
            c = r[k]
            if field:
                n = c.entity.get(field, default)
            else:
                n = c.name
            d[k] = n
    return d


def _data_range(data_array):
    """

    :param data_array:
    :return:
    """
    mx = 0.0
    mn = 0.0
    for i, data in enumerate(data_array):
        for k in accumulate(sorted(data, reverse=True)):
            if k > mx:
                mx = k
            if k < mn:
                mn = k
    return mn, mx


def _res_range(ress):
    """
    accepts a list of results; returns the largest and smallest data values encountered, assuming the segments are
    presented in descending order
    :param ress:
    :return:
    """
    mx = 0.0
    mn = 0.0
    for i, res in enumerate(ress):
        lo = res.total()
        hi = max(res.range())
        if lo < mn:
            mn = lo
        if hi > mx:
            mx = hi
    return mn, mx


class WaterfallChart(object):
    """
    A WaterfallChart turns a collection of LciaResult objects into a collection of waterfall graphs that share an
    ordinal axis. <- this is true, just more correctly interpreted as "aligned abcissas, common ordinate"
    """
    _draw_ylabel = (0,)

    def _stage_style(self, stage):
        if self._names:
            stage = self._names.get(stage, stage)
        if stage in self._color_dict:
            this_style = {'color': self._color_dict[stage]}
        else:
            color = self._color or self._q.get('color') or random_color(self._q.uuid)
            this_style = {'color': color}
        if stage in self._style_dict:
            this_style.update(self._style_dict[stage])
        else:
            if self._style is not None:
                this_style.update(self._style)
        return this_style

    def _create_figure_and_axes(self, size, aspect, case_sep, n_cols):
        """
        Computes the figure size based on the content and creates the axes
        :param size: width of figure in inches
        :param aspect: ratio of bar height to axes width
        :param case_sep: vertical gap between cases in multiples of bar height
        :param n_cols: Number of columns. if None, follow default behavior (one row, unless aspect > 3:1, in which case
        act as though n_cols = 4)
        :return:
        """
        stage_count = sum(len(k) for k in self._stages_by_case.values())
        stage_count += sum(int(k) for k in self._include_net.values())

        # decision point: are we doing a standard cases x quantities single row, or a quantities grid?
        # decide based on the aspect of the final chart.
        # old: height = num_ax * (self._size * aspect * num_steps) + (num_ax - 1) * panel_sep
        wid = 1 / len(self._qs)
        height = (size * wid) * aspect * (stage_count + case_sep * len(self._cases) * int(len(self._cases) > 1))

        # a mess: re-arrange the grid if aspect is too large or if user requests it
        if n_cols or ((size / height) > 3 and len(self._qs) > 4):
            # overrule! do a Kx4 grid of axes
            try:
                n_cols = int(n_cols)
            except TypeError:
                n_cols = 4

            n_cols = min([n_cols, len(self._qs)])

            wid = 1 / n_cols
            n_rows = ceil(len(self._qs) / n_cols)
            row_height = (size * wid) * aspect * (stage_count + case_sep * len(self._cases) * int(len(self._cases) > 1))
            height = row_height * n_rows + self._row_sep * (n_rows - 1)
            ht = 1 / n_rows

            print('Creating %dx%d Layout' % (n_rows, n_cols))
            self._draw_ylabel = tuple(i for i in range(len(self._qs)) if i % n_cols == 0)

            def _ax_pos(_i):
                _lt = wid * (_i % n_cols)
                _bt = ht * (n_rows - 1 - (_i // n_cols))
                return [_lt, _bt, (wid - self._col_sep / size), (ht - self._row_sep / size)]

        else:
            # normal- do case x quantity
            print('Creating 1x%d Layout' % (len(self._qs)))

            def _ax_pos(_i):
                return [wid * _i, 0, (wid - self._col_sep / size), 1]

        self._size = size * wid
        print('Creating figure with size %.2f x %.2f' % (size, height))
        self._fig = plt.figure(figsize=[size, height])

        for i in range(len(self._qs)):

            # ax_pos = [wid * i, 0, (wid - self._col_sep / size), 1]

            ax = self._fig.add_axes(_ax_pos(i))

            ax.invert_yaxis()
            ax.spines['top'].set_visible(False)
            ax.spines['bottom'].set_visible(True)
            ax.spines['bottom'].set_linewidth(2)
            ax.spines['left'].set_visible(False)
            ax.spines['right'].set_visible(False)
            ax.tick_params(labelsize=10)
            ax.tick_params(axis='y', length=0)

            self._ax[self._qs[i]] = ax

    def __init__(self, *results, stages=None, count=None, color=None, color_dict=None,
                 style=None, style_dict=None, stage_property=None, stage_default='Other',
                 include_net=True, net_name='remainder', data_free=False,
                 filename=None, size=6, autorange=False, font_size=None,
                 aspect=0.1, case_sep=None, row_sep=None, col_sep=0.25, n_cols=None, **kwargs):
        """
        Create a waterfall chart that compares the stage contributions of separate LciaResult objects.

        The positional parameters must all be LciaResult objects.  The results should all share the same quantity.

        iterable 'stages' specifies the sequence of queries to the results and the corresponding sequence of waterfall
        steps.  The stage entries are used as an argument to contrib_query. If none is specified, then a collection is
        made of all components of all supplied results.  The collection will have random order.

        The color specification is used for all stages.  Exceptions are stored in color_dict, where the same key
        used for the contrib query, if present, is used to retrieve a custom color for the stage.

        If no default color is specified, a random hue is picked based on the UUID of the quantity.

        Each bar is drawn in the same default color, unless
        :param results: positional parameters must all be LciaResult objects having the same quantity
        :param stages: an ordering of stages. use grab_stages(*results) to get a complete list, then permute it.
        (default ordering is by score of the first quantity, decreasing)
        :param count: [None] limit to the first <count> stages

        :param color: color to use for all bars
        :param color_dict: dictionary of stage name to color for specific stages

        :param style: default style spec for each bar (to override pyplot bar/barh default
        :param style_dict: dict of dicts for custom styles

        :param stage_property: if stages is not specified explicitly, property to grab for stage naming (default
         is to grab entity.name)
        :param stage_default: ['Other'] if stage_property is specified, label to use if the property is missing

        :param include_net: [True] whether to include a net-result bar, if a discrepancy exists between the stage query
         and the total result. This is ignored- a discrepancy is always reported
        :param net_name: ['remainder'] what to call the net-result bar

        :param data_free: [False] whether to suppress printing numeric values

        :param n_cols: number of columns in which to arrange the plots. Defaults to one row, except if n > 4 and the
        resulting figure has an aspect ratio of greater than 3:1, in which case uses a kx4 grid.

        :param filename: default 'waterfall_%.3s.eps' % uuid.  Enter 'none' to return (and not save) the chart
        :param size: axes size in inches (default 6") (width for horiz bars; height for vert bars)
        :param autorange: [False] whether to auto-range the results (no longer supported)
        :param font_size: [None] set text [numbers smaller]
        :param kwargs: aspect: bar height per fig width [0.1]
        case_sep [2.1 bar widths, or 1.2 bar widths if data-free (no cumulative indicator)],
        col_sep=0.25in,
        num_format [%3.2g],
        bar_width [0.85]
        font_size [None]
        row_sep [None]
        """

        self._qs = []
        self._cases = []
        self._res_by_case = defaultdict(list)
        self._res_by_q = defaultdict(list)
        self.data_free = bool(data_free)
        if case_sep is None:
            if self.data_free:
                case_sep = 1.2
            else:
                case_sep = 2.1

        # sort/group results by case and by quantity
        for res in results:
            if res.quantity not in self._qs:
                self._qs.append(res.quantity)
            if res.scenario not in self._cases:
                self._cases.append(res.scenario)
            self._res_by_case[res.scenario].append(res)
            self._res_by_q[res.quantity].append(res)

        # support limited, idiosyncratic customization of styles and colors
        self._color = color  # or self._q.get('color') or random_color(self._q.uuid)
        self._color_dict = color_dict or dict()
        self._style = style or None
        self._style_dict = style_dict or dict()
        self._font_size = font_size
        fontsize = self._font_size or 12

        # group stages by case
        labels = []
        if stages is None:
            stages = {case: grab_stages(*ress, count=count) for case, ress in self._res_by_case.items()}
            self._names = grab_names(*results, field=stage_property, default=stage_default)

        else:
            # I have no idea what's going on here
            stages = {case: list(filter(lambda x: x in grab_stages(*ress, count=count), stages))
                      for case, ress in self._res_by_case.items()}
            self._names = None

        self._stages_by_case = stages

        # prepare net-balance bar (included any time the stage query does not equal the result total
        self._include_net = dict()
        self._net_name = net_name

        # extract data from LciaResult objects-- and compute net-balance
        for case in self._cases:
            if self._names:
                labels += [self._names[k] for k in self._stages_by_case[case]]
            else:
                labels += self._stages_by_case[case]

            net_flag = False
            for res in self._res_by_case[case]:
                data, net = res.contrib_new(*self._stages_by_case[case], autorange=autorange)
                _range = _data_range([data])
                if abs(net) * 1e8 > (_range[1] - _range[0]):
                    # only include remainder if it is greater than 10 ppb
                    net_flag = True
                    labels.append(self._net_name)
                    break

            nf = net_flag and include_net
            self._include_net[case] = nf

        self._row_sep = row_sep or (size / 10)  # real gap in inches between rows
        self._col_sep = col_sep  # real gap in inches between panels
        self._size = 0.0  # real axes width in inches, set in _create_figure()
        self._q = None
        self._ax = dict()
        self._fig = None
        self._span = (0, 0)  # for current axes only

        self._create_figure_and_axes(size, aspect, case_sep, n_cols)

        for i, q in enumerate(self._qs):
            self._q = q
            self._unit = q.unit
            self._span = [k * 1.05 for k in _res_range([k for k in self._res_by_q[q]])]

            ax = self._ax[q]
            ax.set_xlim(self._span)

            start = 0
            yticks = []

            case_res = dict()
            for res in self._res_by_q[q]:
                if res.scenario in case_res:
                    # the proper approach to this is to use multiple entries to draw sensitivity bars
                    raise KeyError('Multiple results for case %s, quantity %s' % (res.scenario, q))
                case_res[res.scenario] = res

            for case in self._cases:
                # write scenario name on the left if there are multiple scenarios
                if i in self._draw_ylabel:
                    if case and case != 'None':  # len(self._cases) > 1:
                        ax.text(0, start - 0.5, '%s ' % case, ha='right', fontsize=fontsize, fontweight='bold')
                start, yt = self._waterfall_case_horiz(ax, case_res[case], start, **kwargs)
                yticks.extend(yt)

                if case != self._cases[-1]:
                    start += case_sep
                else:
                    if self.data_free is False:
                        start += 1.2
                # if len(self._cases) > 1:
                # add scenario name text

            # ## not sure why we are doing this
            # yticks.append(yticks[-1] + case_sep)
            # stgs.append('')

            if 0:  # len(self._cases) == 1:
                # scenario name in title
                sc_name = self._cases[0] or ''
                ax.set_title('%s\n%s' % (self._q['Name'], sc_name), fontsize=fontsize)
            else:
                # scenario name handled above
                ax.set_title('%s\n%s' % (self._q['Name'], self._q.unit), fontsize=fontsize)

                # ## not sure why we are doing this
                # yticks.append(yticks[-1] + case_sep)
                # stgs.append('')

            if i in self._draw_ylabel:
                ax.set_yticks(yticks)
                ax.set_yticklabels(labels)
                ax.set_ylim([start, -1])
            else:
                ax.set_yticks([])
                ax.set_ylim(self._ax[self._qs[0]].get_ylim())

            # vertical axis
            ax.plot([0, 0], ax.get_ylim(), linewidth=2, color=(0, 0, 0), zorder=-1)

            # x labels
            '''
            if abs(cum) > self.int_threshold and abs(mx) > self.int_threshold:
                xticks = [0]
                xticklabels = ['0']
            else:
                xticks = []
                xticklabels = []

            if abs(mx - cum) > self.int_threshold:
                xticks.extend([cum, mx])
                xticklabels.extend([num_format % cum, num_format % mx])
            else:
                xticks.append(cum)
                xticklabels.append(num_format % cum)
                
            '''

            '''
            # add an indicator unit notation to the rightmost tick label
            xticklabels = [_i.get_text() for _i in ax.get_xticklabels()]
            xticks = ax.get_xticks()
            bgst = next(_i for _i in range(len(xticks)) if xticks[_i] == max(xticks))
            xticklabels[bgst] += ' %s' % self._unit
            ax.set_xticks(xticks)
            ax.set_xticklabels(xticklabels)
            # ax.set_xticklabels(xticklabels)
            '''
            ax.set_xlabel(self._q.unit)
            ax.ticklabel_format(axis='x', scilimits=(-3, 3))

            if self.data_free:
                ax.set_xticks([])

            # font size
            if self._font_size:
                for item in ([ax.title, ax.xaxis.label, ax.yaxis.label] + ax.get_xticklabels() + ax.get_yticklabels()):
                    item.set_fontsize(self._font_size)
                t = ax.xaxis.get_offset_text()
                t.set_size(self._font_size)

        if filename != 'none':
            if filename is None:
                uus = '_'.join('%.3s' % q.uuid for q in self._qs)
                filename = 'waterfall-%s.eps' % uus

            save_plot(filename)

    def _waterfall_case_horiz(self, ax, res, start, num_format='%3.2g', bar_width=0.85):
        """

        :param ax:
        :param res:
        :param start:
        :param num_format:
        :param bar_width:
        :return:
        """
        print('Case: %s Quantity: %s ' % (res.scenario, res.quantity))
        _low_gap = 0.25  # this is how much space we want between the bottom bar and the axis?

        fontsize = self._font_size or 12

        cum = 0.0
        center = start + 0.5 - _low_gap
        # top = center - 0.6 * bar_width

        yticks = []

        data, net = res.contrib_new(*self._stages_by_case[res.scenario])  # autorange omitted
        for i, dat in enumerate(data):
            yticks.append(center)

            style = self._stage_style(self._stages_by_case[res.scenario][i])

            if self.data_free:
                self._draw_one_bar_without_label(ax, style, center, cum, dat, bar_width)
            else:
                self._draw_one_bar_with_label(ax, style, center, cum, dat, bar_width, num_format)

            cum += dat
            center += 1

        if self._include_net[res.scenario]:
            yticks.append(center)
            style = {'color': net_color}
            if self.data_free:
                self._draw_one_bar_without_label(ax, style, center, cum, net, bar_width)
            else:
                self._draw_one_bar_with_label(ax, style, center, cum, net, bar_width, num_format)

            cum += net
            center += 1

        if self.data_free:
            # return now
            return center, yticks

        # cumsum marker
        if self._font_size:
            markersize = self._font_size - 2
        else:
            markersize = 8
        ax.plot([cum, cum], [center - 1 + 0.5*bar_width, center], color=(0.3, 0.3, 0.3), zorder=-1, linewidth=0.5)
        ax.plot(cum, center - 0.4 * _low_gap, marker='v', markerfacecolor=(1, 0, 0), markeredgecolor='none',
                markersize=markersize)
        '''
        Cumulative label: we want this to appear below the arrow, but we also want to correct if it is too
        close to the vertical axis or to the edge of the axes.
        
        Multi-case as with the segment labels; unfortunately they are different cases.
        
        First: if the cum is within the threshold of the positive span, then right-align it at the positive span
        Second: if the cum is within the threshold of the negative span, then left-align it at the negative span
        Third: if the cum is within the threshold of 0 on the positive side, then left-align it at 0
        Fourth: if the cum is within the threshold of 0 on the negative side, then right-align it at 0
        Fifth: print it centered under the carat
        
        '''
        tot_thresh = self.int_threshold  # * fontsize / 10
        if cum > self._span[1] - tot_thresh:
            x = self._span[1]
            ha = 'right'
            # cs = 'AAA'
        elif abs(cum) < tot_thresh:
            x = tot_thresh * 0.25
            if cum < 0:
                x *= -1
                ha = 'right'
                # cs = 'CCC'
            else:
                ha = 'left'
                # cs = 'DDD'
        elif cum < self._span[0] + tot_thresh:
            x = self._span[0]
            ha = 'left'
            # cs = 'BBB'
        else:
            # cs = 'EEE'
            x = cum
            ha = 'center'

        total_format = num_format or '%.3g'
        ax.text(x, center + _low_gap, total_format % cum, ha=ha, va='top', fontsize=fontsize)

        return center, yticks

    @staticmethod
    def _draw_one_bar_without_label(ax, style, center, cum, dat, bar_width):
        _conn_color = (0.3, 0.3, 0.3)
        color = style['color']

        if dat < 0:
            style['color'] = _fade_color(color)

        ax.barh(center, dat, left=cum, height=bar_width, **style)

        # connector
        if cum != 0:
            ax.plot([cum, cum], [center - 0.5 * bar_width, center - 1 + 0.5 * bar_width],
                    color=_conn_color, zorder=-1, linewidth=0.5)

    def _draw_one_bar_with_label(self, ax, style, center, cum, dat, bar_width, num_format):
        _h_gap = self.int_threshold * 0.11
        _conn_color = (0.3, 0.3, 0.3)

        label_args = {}
        if self._font_size:
            label_args['fontsize'] = self._font_size * 0.8

        color = style['color']

        if dat < 0:
            style['color'] = _fade_color(color)

        ax.barh(center, dat, left=cum, height=bar_width, **style)
        if num_format:
            if self.int_threshold is not None and abs(dat) > self.int_threshold:
                if sum(style['color'][:2]) < 0.6:
                    text_color = (1, 1, 1)
                else:
                    text_color = (0, 0, 0)

                # interior label
                x = cum + (dat / 2)
                ax.text(x, center, num_format % dat, ha='center', va='center', color=text_color, **label_args)
            else:
                '''# end label positioning-- this is complicated!
                IF the bar is positive and the result is not too far to the right, we want the label on the right
                IF the bar is too far to the right, we want the label on the left regardless of direction
                IF the bar is too far to the left, we want the label on the right regardless of direction
                BUT if the bar is close to 0, we want it printed on the far side from the y axis, to not overwrite
                We know if we're here, the bar is short.  so we only need to think about one end.
                '''
                if cum + dat > self._span[1] - self.int_threshold:
                    # must do left: too close to right
                    if 0 < cum < self.int_threshold:
                        anchor = 'zero left'
                    else:
                        anchor = 'left'
                elif cum + dat < self._span[0] + self.int_threshold:
                    # too close to left
                    if cum < 0 and abs(cum) < self.int_threshold:
                        anchor = 'zero right'
                    else:
                        anchor = 'right'
                elif abs(cum) < self.int_threshold:
                    if cum >= 0:
                        anchor = 'right'
                    else:
                        anchor = 'left'
                else:
                    # not in a danger zone
                    if dat >= 0:
                        anchor = 'right'
                    else:
                        anchor = 'left'

                if anchor == 'left':
                    x = min([cum, cum + dat]) - _h_gap
                    ha = 'right'
                elif anchor == 'zero left':
                    x = -_h_gap
                    ha = 'right'
                elif anchor == 'zero right':
                    x = _h_gap
                    ha = 'left'
                else:
                    x = max([cum, cum + dat]) + _h_gap
                    ha = 'left'

                ax.text(x, center, num_format % dat, ha=ha, va='center', **label_args)

        # connector
        if cum != 0:
            ax.plot([cum, cum], [center - 0.5 * bar_width, center - 1 + 0.5 * bar_width],
                    color=_conn_color, zorder=-1, linewidth=0.5)

    @property
    def fig(self):
        return self._fig

    @property
    def int_threshold(self):
        """
        Useful only for horiz charts
        :return: about 0.5" in axis units (ie. 2x size), scaled up by fontsize / 10pt
        """
        fontsize = self._font_size or 10
        return (self._span[1] - self._span[0]) / (self._size * 2 * 10 / fontsize)

    '''
    def _waterfall_staging_vert(self, scenarios, stages, styles, aspect=0.1, panel_sep=0.75, **kwargs):
        """
        For the vertical-bar waterfall charts, we make just one axes and position the waterfalls at different x
        positions. We may try to do the same thing for horiz waterfallos if we like it better.  This is nice because
        it automatically adjusts for results with

        :param scenarios:
        :param stages:
        :param styles:
        :param aspect:
        :param panel_sep:
        :param kwargs:
        :return:
        """
        num_ax = len(self._d)
        num_steps = len(stages)
        width = num_ax * (self._size * aspect * num_steps) + (num_ax - 1) * panel_sep

        _ax_wid = self._size * aspect * num_steps / width
        _gap_wid = panel_sep / width

        fig = plt.figure(figsize=[width, self._size])
        left = 0.0

        _mn = _mx = 0.0
        axes = []
        for i in range(num_ax):
            right = left + _ax_wid
            ax = fig.add_axes([left, 0.0, _ax_wid, 1.0])
            axes.append(ax)
            self._waterfall_vert(ax, self._d[i], styles, **kwargs)
            ax.set_xticklabels(stages)
            xticklabels = [_i.get_text() for _i in ax.get_xticklabels()]
            xticklabels[-1] += ' %s' % self._q.unit()
            ax.set_xticklabels(xticklabels)

            xlim = ax.get_xlim()
            if xlim[0] < _mn:
                _mn = xlim[0]
            if xlim[1] > _mx:
                _mx = xlim[1]

            if scenarios[i] is not None or num_ax > 1:
                ax.set_title(scenarios[i], fontsize=12)

            top = bottom - _gap_hgt

        for ax in axes:
            ax.set_xlim([_mn, _mx])
    '''

    '''
    def _waterfall_staging_horiz(self, scenarios, stages, styles,
                                 aspect=0.1, panel_sep=0.65,
                                 **kwargs):
        """
        Creates a figure and axes and populates them with waterfalls.
        :param scenarios:
        :param stages:
        :param styles:
        :param aspect:
        :param panel_sep:
        :param kwargs: num_format=%3.2g, bar_width=0.85
        :return:
        """
        num_ax = len(self._d)
        num_steps = sum(sum([k != 0 for k in j]) for j in self._d)  # len(stages)
        height = num_ax * (self._size * aspect * num_steps) + (num_ax - 1) * panel_sep

        _gap_hgt = panel_sep / height

        fig = plt.figure(figsize=[self._size, height])
        top = 1.0

        _mn = _mx = 0.0
        axes = []
        for i in range(num_ax):

            data = [k for k in self._d[i] if k != 0]
            stgs = [stages[j] for j, k in enumerate(self._d[i]) if k != 0]

            _ax_hgt = self._size * aspect * len(data) / height
            bottom = top - _ax_hgt
            ax = fig.add_axes([0.0, bottom, 1.0, _ax_hgt])
            axes.append(ax)
            self._waterfall_horiz(ax, data, styles, **kwargs)
            ax.set_yticklabels(stgs)


            # add an indicator unit notation to the rightmost tick label
            xticklabels = [_i.get_text() for _i in ax.get_xticklabels()]
            xticks = ax.get_xticks()
            bgst = next(_i for _i in range(len(xticks)) if xticks[_i] == max(xticks))
            xticklabels[bgst] += ' %s' % self._unit
            ax.set_xticklabels(xticklabels)


            xlim = ax.get_xlim()
            if xlim[0] < _mn:
                _mn = xlim[0]
            if xlim[1] > _mx:
                _mx = xlim[1]

            if scenarios[i] is not None or num_ax > 1:
                sc_name = scenarios[i]
            else:
                sc_name = ''

            fontsize = self._font_size or 12

            if i == 0:
                ax.set_title('%s [%s]\n%s' % (self._q['Name'], self._unit, sc_name), fontsize=fontsize)
            else:
                ax.set_title('%s' % sc_name, fontsize=fontsize)

            top = bottom - _gap_hgt

        for ax in axes:
            ax.set_xlim([_mn, _mx])
            # ax.set_xticks(ax.get_xticks() + [_mx])
            # ax.set_xticklabels(ax.get_xticklabels() + [str(self._unit)])
        return fig
    '''

    '''
    def _waterfall_horiz(self, ax, num_format='%3.2g', bar_width=0.85):
        """

        :param ax:
        :param data:
        :param styles: a list of style kwargs to add to each bar.  must include 'color' key; all others extra.
        :param num_format:
        :param bar_width:
        :return:
        """
        """
        The axes are already drawn. all we want to do is make and label the bars, one at a time.
        """
        _low_gap = 0.25

        _conn_color = (0.3, 0.3, 0.3)

        cum = 0.0
        center = 0.5 - _low_gap
        top = center - 0.6 * bar_width
        bottom = floor(center + len(data))

        # vertical axis
        ax.plot([0, 0], [top, bottom], linewidth=2, color=(0, 0, 0), zorder=-1)

        _h_gap = self.int_threshold * 0.11
        yticks = []

        mx = 0.0

        label_args = {}
        if self._font_size:
            label_args['fontsize'] = self._font_size - 2

        for i, dat in enumerate(data):
            yticks.append(center)
            style = styles[i]
            self._draw_one_bar_with_label(ax, style, center, cum, dat, bar_width, num_format)
            
            cum += dat
            if cum > mx:
                mx = cum

            center += 1
        ax.invert_yaxis()
        ax.spines['top'].set_visible(False)
        ax.spines['bottom'].set_visible(True)
        ax.spines['bottom'].set_linewidth(2)
        ax.spines['left'].set_visible(False)
        ax.spines['right'].set_visible(False)
        ax.set_yticks(yticks)
        ax.tick_params(labelsize=10)
        ax.tick_params(axis='y', length=0)

        # cumsum marker
        if self._font_size:
            markersize = self._font_size - 2
        else:
            markersize = 8
        ax.plot([cum, cum], [center - 1 + 0.5*bar_width, bottom], color=_conn_color, zorder=-1, linewidth=0.5)
        ax.plot(cum, bottom - 0.4 * _low_gap, marker='v', markerfacecolor=(1, 0, 0), markeredgecolor='none',
                markersize=markersize)

        # x labels
        if abs(cum) > self.int_threshold and abs(mx) > self.int_threshold:
            xticks = [0]
            xticklabels = ['0']
        else:
            xticks = []
            xticklabels = []

        if abs(mx - cum) > self.int_threshold:
            xticks.extend([cum, mx])
            xticklabels.extend([num_format % cum, num_format % mx])
        else:
            xticks.append(cum)
            xticklabels.append(num_format % cum)

        ax.set_xticks(xticks)
        ax.set_xticklabels(xticklabels)

        ### font size
        if self._font_size:
            for item in ([ax.title, ax.xaxis.label, ax.yaxis.label] + ax.get_xticklabels() + ax.get_yticklabels()):
                item.set_fontsize(self._font_size)
    '''



"""
A Qdb interface provider, implemented for now as a google sheet
"""

from xlstools.google_sheet_reader import GoogleSheetReader
from antelope import BaseEntity


INDEX_HEADINGS = ('external_ref', 'Name', 'ShortName', 'Method', 'Category', 'Indicator', 'unit', 'Comment', 'Notes')
FACTOR_HEADINGS = ('flowable', 'context', 'ref_quantity', 'ref_unit', 'locale', 'value')


class QdbQuantity(BaseEntity):
    is_entity = True
    entity_type = 'quantity'

    def __init__(self, origin, external_ref, unit, **kwargs):
        self._origin = origin
        self._external_ref = external_ref
        self._unit = unit
        self._args = kwargs

    @property
    def origin(self):
        """
        Must return a resolvable unique ID, nominally 'origin/external_ref'
        :return:
        """
        return self._origin

    @property
    def external_ref(self):
        return self._external_ref

    @property
    def reference_entity(self):
        """
        Must have a .unit property that returns a string,
        should have .entity_type property that returns 'quantity'
        :return:
        """
        return self._unit

    def properties(self):
        for k in sorted(self._args.keys()):
            yield k

    def get(self, item):
        return self._args.get(item)

    def make_ref(self, query):
        """
        if is_entity is true, entity must return a ref made from the provided query
        :param query:
        :return:
        """
        return NotImplemented


class QdbGSheetClient(object):
    def __init__(self, source, ref, json_credential=None, credential_file=None, **kwargs):
        self._source = source
        self._ref = ref
        cred = json_credential or credential_file
        self._xls = GoogleSheetReader(cred, source)
        self._args = kwargs
        self._qsheet = None
        self._qs = []
        self._qd = dict()
        self.load_all()

    @property
    def static(self):
        return False

    def load_all(self):
        if 'Quantities' not in self._xls.sheet_names():
            self._xls.create_sheet('Quantities')
            self._xls.write_row('Quantities', 0, INDEX_HEADINGS)
        self._qsheet = self._xls.sheet_by_name('Quantities')
        self._qs = [self._qsheet.row_dict(i) for i in range(1, self._qsheet.nrows)]
        self._qd = {k['external_ref']: k for k in self._qs}



import logging
try:
    from xlstools.google_sheet_reader import GoogleSheetReader
except ImportError:
    logging.error('No GoogleSheetReader found')

    class GoogleSheetReader:
        def __init__(self, credentials, spreadsheet_id):
            self.credentials = credentials
            self.sheet_id = spreadsheet_id

from .observer.headers import HEADERS, WIDTHS, TAB_COLORS, SpannerMeta


class GSheetObserver(GoogleSheetReader):

    def _request_text_format_row(self, sheet, row, start_column=0, **kwargs):
        sheet_id = self.sheet_id(sheet)
        return {
            'repeatCell': {
                'range': {
                    'sheetId': sheet_id,
                    'startRowIndex': row,
                    'endRowIndex': row + 1,
                    'startColumnIndex': start_column
                },
                'cell': {
                    'userEnteredFormat': {
                        'textFormat': {
                            **kwargs
                        }
                    }
                },
                'fields': 'userEnteredFormat(textFormat)'
            }
        }

    def _request_column_width(self, sheet, column, pixel_size=120, num_columns=1, **kwargs):
        sheet_id = self.sheet_id(sheet)
        return {
            'updateDimensionProperties': {
                'range': {
                    'sheetId': sheet_id,
                    'dimension': 'COLUMNS',
                    'startIndex': column,
                    'endIndex': column + num_columns
                },
                'properties': {
                    'pixelSize': pixel_size,
                    **kwargs
                },
                'fields': ','.join(['pixelSize'] + list(kwargs.keys()))
            }
        }

    def _request_sheet_properties(self, sheet, col_freeze=None, row_freeze=None, **kwargs):
        sheet_id = self.sheet_id(sheet)
        grid_props = dict()
        fields = list(kwargs.keys())
        if col_freeze:
            grid_props['frozenColumnCount'] = int(col_freeze)
            fields.append('gridProperties.frozenColumnCount')
        if row_freeze:
            grid_props['frozenRowCount'] = int(row_freeze)
            fields.append('gridProperties.frozenRowCount')
        return {
            'updateSheetProperties': {
                'properties': {
                    'sheetId': sheet_id,
                    'gridProperties': grid_props,
                    **kwargs
                },
                'fields': ','.join(fields)
            }
        }

    def _request_data_validation_column(self, sheet, column, condition, num_cols=1, start_row=0):
        sheet_id = self.sheet_id(sheet)
        return {
            'setDataValidation': {
                'range': {
                    'sheetId': sheet_id,
                    'startRowIndex': start_row,
                    'startColumnIndex': column,
                    'endColumnIndex': column + num_cols
                },
                'rule': {
                    'condition': condition,
                    'inputMessage': 'select a flow',
                    'strict': False,
                    'showCustomUi': True
                }
            }
        }

    def _create_and_format_sheet_by_recipe(self, recipe, sheet_name=None):
        if sheet_name is None:
            sheet_name = recipe
        self.create_sheet(sheet_name)
        sheet_id = self.sheet_id(sheet_name)

        h = HEADERS[recipe]
        ws = WIDTHS.get(recipe, dict())
        tab_color = TAB_COLORS.get(recipe)
        rgb = {'red': tab_color[0], 'green': tab_color[1], 'blue': tab_color[2]}

        self.write_row(sheet_name, 0, h)
        self.batch_update(self._request_text_format_row(sheet_name, 0, bold=True),
                          self._request_sheet_properties(sheet_name, tabColorStyle={'rgbColor': rgb}),
                          *(self._request_column_width(sheet_name, h.index(i), w) for i, w in ws.items()))

        return sheet_id

    def _production_sheet_config(self, sheet_name):
        condition = {
            'type': 'ONE_OF_RANGE',
            'values': [
                {'userEnteredValue': '=flows!$A:$A'}
            ]
        }
        self.batch_update(self._request_sheet_properties(sheet_name, col_freeze=4, row_freeze=1),
                          self._request_data_validation_column(sheet_name,
                                                               HEADERS['production'].index('child_flow'),
                                                               condition,
                                                               start_row=1))

    def build_observatory(self):
        """
        Initializes the 'quantities', 'flows', 'flowproperties', 'spanners', 'production', and 'observations' sheets
        in the supplied xlrd-like.  (note: currently only works for GoogleSheetReader objects from xlstools)

        sets the headers to bold on all sheets; sets column widths and tab colors

        on the 'production' sheet, freezes 4 columns and 1 row, sets data validation on "child_flow"

        :return:
        """
        for k in ('quantities', 'flows', 'flowproperties', 'production', 'observations', 'taps', 'spanners'):
            if k not in self._sheetnames:
                # only create the sheet if it doesn't already exist
                self._create_and_format_sheet_by_recipe(k)

                if k == 'production':
                    self._production_sheet_config(k)

    def new_production_sheet(self, sheet_name='production'):
        if sheet_name not in self._sheetnames:
            # only create the sheet if it doesn't already exist
            self._create_and_format_sheet_by_recipe('production', sheet_name=sheet_name)

        self._production_sheet_config(sheet_name)

    @property
    def spanners(self):
        """
        Generates spanners by entry in 'spanners'
        :return:
        """
        spanners_meta = self.sheet_by_name('spanners')
        for n in range(1, spanners_meta.nrows):
            d = spanners_meta.row_dict(n)
            if d['external_ref'] is None:
                continue
            if d['external_ref'] in self._sheetnames:
                yield SpannerMeta.from_form(**d)

    def new_spanner(self, spanner_ref, source=None, spanners='spanners', **kwargs):
        # validates ref
        spanner_model = SpannerMeta.from_form(spanner_ref, source=source or 'gsheet', **kwargs)
        spanner_data = spanner_model.model_dump()

        if spanner_ref not in self._sheetnames:
            self._create_and_format_sheet_by_recipe('spanner', sheet_name=spanner_ref)
            self.batch_update(self._request_sheet_properties(spanner_ref, col_freeze=1, row_freeze=1))

        spanners_meta = self.sheet_by_name(spanners)
        rec = [k.value for k in spanners_meta.col(0)]
        try:
            row = rec.index(spanner_ref)
        except ValueError:
            row = spanners_meta.nrows

        self.write_row(spanners, row, [spanner_data.get(k) for k in HEADERS['spanners']])
        return spanner_model



from .lc_mfa_study import NestedLcaStudy
from typing import Dict, Tuple, Optional
from pydantic import BaseModel

from antelope import comp_dir, EntityNotFound


class StudySpec(BaseModel):

    stage_names: Dict[str, str] = dict()  # target: stagename
    logistics_mappings: Dict[str, Tuple[str, str]] = dict()  # flow: target, stagename
    activity_mappings: Dict[Tuple[str, str], Tuple[str, str]] = dict()  # direction, flow: target, stagename
    supply_routes: Dict[str, Tuple] = dict()  # name: tuple-spec, sense='Source'
    disposition_routes: Dict[str, Tuple] = dict()  # name: tuple-spec, sense='Sink'
    study_sinks: Dict[str, Dict[str, Optional[float]]] = dict()  # flow: {target: share}
    study_sources: Dict[str, Dict[str, Optional[float]]] = dict()  # flow: {target: share}


class ObservedMfaStudy(NestedLcaStudy):
    def make_study_mapping(self, container, flow, direction, target, stage=None, scenario=None):
        if stage is None:
            stage = container['Name']
        f = self._resolve_term(flow)
        try:
            cf = next(container.children_with_flow(f, direction=direction))
        except StopIteration:
            cf = self.fg.new_fragment(f, direction, parent=container)

        term = self._resolve_term(target)
        cf.clear_termination(scenario=scenario)
        cf.terminate(term, scenario=scenario, descend=False)
        if stage:
            cf['StageName'] = stage
        return cf

    '''
    def add_logistics_route(self, flow, provider, descend=False, **kwargs):
        try:
            return next(self.logistics_container.children_with_flow(flow))
        except StopIteration:
            c = self.fg.new_fragment(flow, 'Input', parent=self.logistics_container, **kwargs)
            c.terminate(self._resolve_term(provider), descend=descend)
            return c
    '''

    def make_logistics_mappings(self, logistics_mappings, **kwargs):
        for k, v in logistics_mappings.items():
            t, stage = v
            self.make_study_mapping(self.logistics_container, k, 'Input', t, stage=stage, **kwargs)

    def make_activity_mappings(self, activity_mappings):
        for k, v in activity_mappings.items():
            d, f = k
            t, s = v
            self.make_study_mapping(self.activity_container, f, d, t, stage=s)

    def _make_study_market(self, flow_ref, sense, market_spec, stage_names=None):
        direction = comp_dir(sense)
        try:
            flow = self.data.get(flow_ref)
        except EntityNotFound:
            flow = self._resolve_term(flow_ref)
        try:
            next(self.study_container.children_with_flow(flow, direction=direction))
        except StopIteration:
            mkt = self.fg.new_fragment(flow, direction, parent=self.study_container)
            self.make_market(mkt, market_spec, sense, stage_names=stage_names)

    def make_study_sources(self, study_sources, stage_names=None):
        for k, v in study_sources.items():
            self._make_study_market(k, 'Source', v, stage_names)

    def make_study_sinks(self, study_sinks, stage_names=None):
        for k, v in study_sinks.items():
            self._make_study_market(k, 'Sink', v, stage_names)

    def make_study(self, study_spec: StudySpec):
        # study building blocks
        self.make_routes(study_spec.supply_routes, sense='Source', stage_names=study_spec.stage_names)
        self.make_routes(study_spec.disposition_routes, sense='Sink', stage_names=study_spec.stage_names)

        # transformation activities
        self.make_activity_mappings(study_spec.activity_mappings)

        # logistics activities
        self.make_logistics_mappings(study_spec.logistics_mappings)

        # terminations for study inflows and outflows
        self.make_study_sources(study_spec.study_sources, stage_names=study_spec.stage_names)
        self.make_study_sinks(study_spec.study_sinks, stage_names=study_spec.stage_names)



q
from .base import BaseTableOutput
from collections import defaultdict
from antelope_core.exchanges import ExchangeValue


class AllocationGrid(BaseTableOutput):
    """
    This table can be produced dynamically

    """
    _near_headings = 'Direction', 'Ref'  # should be overridden
    _returns_sets = True
    _far_headings = 'Flow', 'Comment'

    def _generate_items(self, alloc_inv):
        """
        columns should be seeded with allocated exchanges

        :param alloc_inv: the inventory iterable
        :return: allocated exchanges
        """
        for x in alloc_inv:
            if not x.is_reference:
                yield x

    def _pull_row_from_item(self, item):
        """
        :param item: an allocated exchange
        :return: 3-tuple: direction, is_ref (bool), flow name, flow compartment
        """
        row = item.direction, False, '; '.join(item.flow.context), item.flow['Name']
        if row not in self._notes:
            self._notes[row] = self._pull_note_from_item(item)
        return row

    def _pull_note_from_item(self, item):
        return item.comment

    def _canonical(self, item):
        """
        canonical row name for entry
        :param item:
        :return:
        """
        term = item.termination
        if term is not None:
            term = self._ar.get(item.termination)
        if term is None:
            return '%s %s' % (item.unit, item.flow)
        else:
            return '%s %s' % (item.unit, term.name)

    def _extract_data_from_item(self, objects):
        """
        Each item here is an exchange. The sets are multiple terminations of the same flow.
        :param objects:
        :return:
        """
        d = defaultdict(float)
        if objects is None:
            return d
        for item in objects:
            value = item.value
            if value is None:
                continue
            d[self._canonical(item)] += value
        return d

    def _add_alloc_column(self, col_idx, arg, ref):
        """
        extracts items; only adds a column if there is a nonzero count of items
        :param arg:
        :param ref:
        :return:
        """
        count = False
        for k in self._generate_items(arg.inventory(ref_flow=ref)):
            if k.value is not None:
                count = True
                self._add_rowitem(col_idx, k)
        return count

    def _add_alloc_refs(self, arg, flow=None):
        col_idx = len(self._columns)
        for k in arg.references(flow=flow):
            rx = ExchangeValue(k.process, k.flow, k.direction, value=1.0)  # we do this to avoid calling RxRef.value
            # (and because table exchanges are normalized to this value, so 1.0 is the only correct value to report)
            rx.set_ref(k.process)
            row = k.direction, True, '; '.join(k.flow.context), k.flow['Name']
            self._add_rowitem(col_idx, rx, row=row)
        self._columns.append(arg)

    def _sorted_rows(self):
        for row in sorted(self._rows, key=lambda x: (not x[1], x[0], x[2], x[3])):
            yield row

    def _build_near_header(self, row, prev):
        the_row = []
        if prev is not None:
            if row[0] == prev[0]:
                the_row.append('""')
            else:
                the_row.append(row[0])
        else:
            the_row.append(row[0])
        if row[1]:
            the_row.append('{*}')
        else:
            the_row.append('')
        return the_row

    def add_column(self, arg):
        col_idx = len(self._columns)
        # add process's native values
        if self._report_unallocated:
            if self._add_alloc_column(col_idx, arg, None):
                self._add_alloc_refs(arg)

        for ref in arg.references():
            col_idx = len(self._columns)
            if self._add_alloc_column(col_idx, arg, ref.flow):
                self._add_alloc_refs(arg, flow=ref.flow)

    def __init__(self, archive, *prefs, report_unallocated=True):
        """

        :param archive:
        :param prefs: one or more process refs
        :param report_unallocated: [True] by default, report process's unallocated values alongside the allocated ones
        """
        self._ar = archive
        self._report_unallocated = bool(report_unallocated)
        super(AllocationGrid, self).__init__(*prefs)

    def to_excel(self, xl_writer, sheetname, width_scaling=0.75):
        super(AllocationGrid, self).to_excel(xl_writer, sheetname, width_scaling=width_scaling)
        ix = self._near_headings.index('Ref') + 1
        center_ref = xl_writer.book.add_format({'align': 'center', 'bold': True})
        xl_writer.sheets[sheetname].set_column(ix, ix, cell_format=center_ref)



from xlstools import GoogleSheetReader
from pandas import to_numeric


def prepare_dataframe(df):
    return df.apply(to_numeric).sort_values(df.columns[0])


class GSheetTableWriter(object):
    _gsheet = None
    _active = False

    def _check_or_create_sheet(self, sheetname, group=None):
        sn = self._gsheet.sheet_names()
        if sheetname in sn:
            ix = sn.index(sheetname)
        else:
            self._gsheet.create_sheet(sheetname)
            ix = len(sn)
        if self._index:
            self._gsheet.write_row('Index', ix, (sheetname, group))

    def write_results_dataframe_to_gsheet(self, sheetname, df, group=None,
                                          clear_sheet=True, write_header=True, header_levels=None):
        """

        :param gsheet: a GoogleSheetReader
        :param sheetname: sheet to write to or create
        :param df: sorted, prepared data
        :param group: [None] optional - reported in Index sheet
        :param clear_sheet: [True]
        :param write_header: [True] whether to write header (False: leave it standing)
        :param header_levels: number of header levels to write. Must be <= nlevels
        :return:
        """

        ncol = len(df.columns)
        if header_levels is None or header_levels > df.columns.nlevels:
            header_levels = df.columns.nlevels

        if sheetname in self._gsheet.sheet_names():
            # start by clearing the sheet- with or without headers
            if clear_sheet:
                if write_header:
                    # we are clearing the entire sheet, to write both header and data
                    self._gsheet.clear_region(sheetname)
                else:
                    # we clearing sheet but keeping header
                    self._gsheet.clear_region(sheetname, start_row=header_levels)
            else:
                # we are not clearing the sheet
                if write_header:
                    # but we are still rewriting the header
                    self._gsheet.clear_region(sheetname, end_col=ncol, end_row=header_levels - 1)
        else:
            self._check_or_create_sheet(sheetname, group=group)
            write_header = True  # force header write if the sheet is just created

        # then populate
        def _row_gen(_df):
            for j, row in _df.fillna('NA').iterrows():
                yield [j] + list(row.values)

        if write_header:
            for i in range(header_levels):
                h = [''] + list(df.columns.get_level_values(i))
                self._gsheet.write_row(sheetname, i, h)
        self._gsheet.write_rectangle_by_rows(sheetname, _row_gen(df), start_row=header_levels)

    @classmethod
    def from_credentials(cls, credentials, gsheet_id, **kwargs):
        return cls(GoogleSheetReader(credentials, gsheet_id), **kwargs)

    def _re_flow_index_sheet(self):
        """
        Re-order rows in the index sheet to match the actual order of tabs.
        :return:
        """
        sn = self._gsheet.sheet_names()
        cur_ix = self._gsheet.sheet_by_name('Index')
        meta = cur_ix.col(0)
        dd = {k.value: [j.value for j in cur_ix.row(i)] for i, k in enumerate(meta) if k.value is not None}

        if cur_ix.nrows > 1:
            self._gsheet.clear_region('Index', start_row=1)
        self._gsheet.write_rectangle_by_rows('Index', (dd[k] for k in sn if k in dd and k != 'Index'), start_row=1)

    def __init__(self, sheet=None, active=None, index=True):
        self._index = index
        self.active = active
        if sheet:
            if active is None:  # if sheet provided and active is not provided, assume activate
                self.active = True
            self._gsheet = sheet
            if self._index:
                # could do some machinations here to detect + manage deleted or re-ordered sheets
                if 'Index' not in sheet.sheet_names():
                    # for now, we just create an index- and leave the description field to manual edit
                    sheet.create_sheet('Index')
                    sheet.write_row('Index', 0, ('Sheet Name', 'Group', 'Description'))
                self._re_flow_index_sheet()

    @property
    def active(self):
        return self._active

    @active.setter
    def active(self, value):
        self._active = bool(value)

    def _write(self, sheet, df, **kwargs):
        if self.active:
            self.write_results_dataframe_to_gsheet(sheet, df, **kwargs)
        else:
            print('[GSheet Write suppressed] %s' % sheet)

    def write_subgroup_results(self, group, runner, scenario, case=None):
        runner.format = None
        df = prepare_dataframe(runner.scenario_detail_tbl(scenario))
        sheetname = '%s-%s' % (group, scenario)
        if case:
            sheetname = '-'.join([sheetname, case])
        self._write(sheetname, df, header_levels=2, group=group)

    def write_detailed_results(self, runner, scenario, group=None):
        runner.format = None
        df = prepare_dataframe(runner.scenario_detail_tbl(scenario))
        self._write(scenario, df, header_levels=2, group=group)

    def write_summary_results(self, runner, summary, group=None):
        runner.format = None
        df = prepare_dataframe(runner.scenario_summary_tbl())
        self._write(summary, df, header_levels=2, group=group)



"""
An object lesson in what happens when you don't learn about existing tools (of the vintage: jinja2)

This file [was] used to cast a fragment to TeX.  The output is a TeX file in an include folder which draws
a fragment as a tree diagram using PSTricks.

I guess for research on this, I should start by actually drawing a fragment by hand
"""
import os
import re
from antelope import comp_dir
from .charts import scenario_compare_figure, save_plot


uuid_regex = re.compile('([0-9a-f]{8}.?([0-9a-f]{4}.?){3}[0-9a-f]{12})')


class TraversalError(Exception):
    pass


default_tex_folder = 'tex-files'
default_frag_data = 'fragment-data.tex'
default_model_doc = 'model-doc.tex'


def tex_sanitize(tex):
    try:
        tex = re.sub('%', '\\%', tex)
        tex = re.sub('_', '\\\\textunderscore{}', tex)  # this doesn't work bc filenames have underscores
    except TypeError:
        tex = tex.__class__.__name__
    return tex


def frag_link_label(name):
    if bool(uuid_regex.match(name)):
        return name[:5]
    else:
        return tex_sanitize(name)


def grab_stages(results, sort=None):
    stages = set()

    def _first_score(_comp):
        for r in results:
            try:
                return r[_comp].cumulative_result
            except KeyError:
                continue
        return 0

    for r in results:
        if isinstance(r, list):
            for k in r:
                stages = stages.union(k.component_entities())
        else:
            stages = stages.union(r.component_entities())

    if sort is None:
        return sorted(stages, key=_first_score, reverse=True)
    else:
        return sorted(stages, key=sort)


"""
drawing 'primitives'
"""
TAB_LF = '\\\\'


def stage_name_table(stages):
    out = '\\scriptsize\n\\begin{tabular}[b]{c@{ -- }l}\n'
    for i, k in enumerate(stages):
        out += '%s & %s\\\\\n' % (chr(ord('A') + i), tex_sanitize(k))
    out += '\\end{tabular}\n'
    return out


def save_stages(fname, stages):
    out = stage_name_table(stages)
    with open(fname + '_stages', 'w') as fp:
        fp.write(out)


def fragment_header(frag):
    s = '{\\hypertarget{%.5s}{\\Large \\texttt{%s}}}\\subsection{%s}\n' % (
        frag.uuid, frag_link_label(frag.external_ref), frag['Name']
    )
    return tex_sanitize(s)


def fragment_fu(frag, scenario=None):
    s = '{\\large %s: %g %s %s}\\\\[8pt]\n%s\n' % (comp_dir(frag.direction), frag.exchange_value(scenario),
        frag.flow.unit,
        frag.flow['Name'], frag['Comment'])
    return tex_sanitize(s)


def fragment_inventory(fragment, scenario=None):
    exchs = [x for x in fragment.cutoffs(scenario=scenario, observed=True)]
    inventory = ''
    if len(exchs) > 0:
        inventory += '\n{\\small\n\\begin{tabular}{rccl}\n'
        for x in exchs:
            inventory += tex_sanitize('%6s & %6.3g & %s & %s\\\\\n' % (x.direction,
                                                                       x.value, x.flow.unit, x.flow['Name']))
        inventory += '\\end{tabular}\n}\n'

    return inventory


def frag_drawing_opener(max_size):
    return '''
\\begin{pspicture}(0,1)(17,%f)\\showgrid
  \\small\\sffamily
  \\rput(0,0.5){
    \\psset{xunit=1.2cm,yunit=1cm,style=standard}''' % (-max_size)


def frag_drawing_closer():
    return '  }\n\\end{pspicture}\n'


def frag_pnodes(coords):
    pnodes = ''
    for c in coords:
        pnodes += '\\pnode(%f,%f){px%s}\n' % (c[1], -c[2], c[0])
    return pnodes


def cutoff_box(uid):
    return '\\rput(px%.5s){\\rput[l](1,0){\\rnode[style=phantom]{nx%.5s}{\scriptsize CUTOFF}}}' % (uid, uid)


def bg_box(uid):
    return '\\rput(px%.5s){\\pnode(1.1,0){nx%.5s}\\psframe*(1,-0.25)(1.1,0.25)}' % (uid, uid)


def subfrag_box(uid, tgt_top):
    return ('\\rput(px%.5s){\\rput(1.25,0){\\nodebox{nx%.5s}{style=process,style=dashed}'
            '{1.15cm}{0.6cm}{\\hyperlink{%.5s}{}}}}'
            '\\rput[b]([angle=90,nodesep=3pt]nx%.5s){\\texttt{%s}}'% (uid, uid, tgt_top.uuid, uid,
                                                                      frag_link_label(tgt_top.external_ref)))


def fg_box(uid):
    return '\\rput(px%.5s){\\rput(1,0){\\circlenode{nx%.5s}{}}}' % (uid, uid)


def io_box(uid):
    return '\\rput(px%.5s){\\pnode(1,0){nx%.5s}}\\psdots[dotsize=4pt](nx%.5s)' % (uid, uid, uid)


def process_box(uid):
    return '\\rput(px%.5s){\\rput(1,0){\\nodebox{nx%.5s}{style=process}{0.8cm}{0.2cm}{}}}' % (uid, uid)


def context_box(uid):
    return '\\rput(px%.5s){\\rput(1,0){\\nodebox{nx%.5s}{style=phantom}{1em}{1em}{$\odot$}}}' % (uid, uid)


class TeXAuthor(object):
    """
    We create a TeXAuthor object to write TeX files about fragments.

    """
    @property
    def img_folder(self):
        return os.path.join(self.folder, 'img')

    @staticmethod
    def _img_fname(f_base):
        return '%s_img.eps' % f_base

    @staticmethod
    def _stg_fname(f_base):
        return '%s_stages' % f_base

    @property
    def _wrapper_fname(self):
        return os.path.join(self.folder, self._fragment_data_file)

    @property
    def _documentation_fname(self):
        return os.path.join(self.folder, self._model_doc_file)

    @staticmethod
    def _init_report_file(fname, intro_text):
        if os.path.exists(fname):
            os.remove(fname)

        with open(fname, 'w') as fp:
            fp.write('%% %s\n' % intro_text)

    def _read_wrapper(self):
        with open(self._wrapper_fname) as fp:
            return fp.read().splitlines()

    def img_rel_path(self, f_base):
        return os.path.join(self.img_folder, self._img_fname(f_base))

    def stg_rel_path(self, f_base):
        return os.path.join(self.img_folder, self._stg_fname(f_base))

    def _write_stage_names(self, f_base, stages):
        out = stage_name_table(stages)
        with open(self.stg_rel_path(f_base), 'w') as fp:
            fp.write(out)

    def __init__(self, folder=default_tex_folder, overwrite=True, comments=False,
                 frag_data=default_frag_data, model_doc=default_model_doc):
        """
        This function

        :param folder: ['tex-files'] Use different folders for different reports if the reports represent different
         scenarios or data.
        :param overwrite: [True] whether to overwrite existing files in the folder
        :param comments: [False] whether to write FragmentFlow comments on tree drawings
        :param frag_data: ['fragment-data.tex'] default fragment data file.  Use the same folder but different frag_data
         files for different reports if the reports are based on the same scenario but include different contents.
        :param model_doc: ['model-doc.tex'] model doc file.
        """
        self.folder = folder
        self.overwrite = overwrite
        self.comments = comments

        self._fragment_data_file = frag_data
        self._model_doc_file = model_doc

        self._seen = set()

        if os.path.exists(folder):
            if not os.path.isdir(folder):
                raise ValueError('Path is not a directory.')
        else:
            os.makedirs(folder)
        if not os.path.exists(self.img_folder):
            os.makedirs(self.img_folder)

        self._file_list = []
        if os.path.exists(self._wrapper_fname):
            for file in self._read_wrapper():
                try:
                    fname = re.search('\\\input\{(.*/)?(.*).tex\}', file).group(2)
                except AttributeError:
                    continue
                if fname not in self._file_list:
                    self._file_list.append(fname)

    def _write_output(self, f_base, content, add_to_index=None):
        """
        Writes output to a specified file basename.  The output is written to f_base.tex, and any image and
        stage data should be generated in img_rel_path(f_base) and stg_rel_path(f_base) respectively.
        :param f_base:
        :param content:
        :param add_to_index: can be None, 'tex', 'doc': whether to add the filename to the tex-files or
         model-documentation index
        :return:
        """
        filename = os.path.join(self.folder, '%s.tex' % f_base)

        with open(filename, 'w') as fp:
            fp.write(content)

        print('Written to %s' % filename)
        if add_to_index == 'tex':
            if f_base not in self._file_list:
                with open(self._wrapper_fname, 'a') as wp:
                    wp.write('\\input{%s}\\clearpage\n' % filename)
                self._file_list.append(f_base)
        elif add_to_index == 'doc':
            with open(self._documentation_fname, 'a') as dp:
                dp.write('\\input{%s}\n' % filename)

    def frag_layout_recurse(self, fragment, level=0, depth=0, scenario=None):
        """
        This pass creates a list of coordinates for pnode declarations for the reference points of all fragment flows.
        returns a 3-tuple of (pnode name, x-coord, y-coord)

        At the end of it, coords[-1][2] is the vertical position of the last child

        Each pnode's name is equal to the first 5 characters of the fragment's uuid, same as in the drawing.
        :param fragment:
        :param level: [0] vertical traversal height, used for positioning the pnodes
        :param depth: [0] horizontal traversal depth
        :param scenario:
        :return:
xs        """
        coords = [('%.5s' % fragment.uuid, depth, level)]
        if fragment.exchange_value(scenario, observed=True) == 0:
            return coords

        term = fragment.termination(scenario)

        children = [c for c in fragment.child_flows]
        if len(children) > 0 and not term.is_null:
            # setup coords for first child
            depth += 2.0
            if term.is_subfrag:
                depth += .5
            level += 1
            for c in sorted(children, key=lambda x: (x['StageName'],
                                                     x.direction,
                                                     not x.termination(scenario).is_null,
                                                     x.termination(scenario).is_bg)):
                c_term = c.termination(scenario)
                #if c.exchange_value(scenario, observed=True) == 0 and not c.is_balance:
                #    continue
                if c_term.is_subfrag:
                    level += .25
                coords.extend(self.frag_layout_recurse(c, level=level, depth=depth, scenario=scenario))
                level = coords[-1][2]
                # setup coords for next child -- background nodes are smaller
                if c_term.is_bg or c_term.is_fg or c_term.is_null:
                    level += 0.7
                else:
                    level += 1

        return coords

    def frag_traversal_entry(self, fragment, scenario=None):
        return self.frag_layout_traverse(fragment, 1.0, scenario=scenario, first=True)

    def frag_layout_traverse(self, fragment, node_weight, scenario=None, first=False, parbox_width=15):
        """
        This actually draws the fragment components
        :param fragment:
        :param node_weight:
        :param scenario:
        :param first: [False] whether to suppress printing exchange weight
        :param parbox_width: [14cm] width of right-hand text box
        :return: boxes, arrows, subfragments
        """
        subfrag_scale = 1.0
        subfrags = []
        term = fragment.termination(scenario)

        if not first:
            node_weight *= fragment.exchange_value(scenario, observed=True)
        #if node_weight == 0 and not fragment.is_balance:
        #    return '\n', '\n', subfrags
        if term.is_frag:
            if term.is_fg:
                # foreground
                boxes = fg_box(fragment.uuid)
                frag_name = fragment['Name']
            elif term.term_is_bg:
                if term.term_node.term.is_null:
                    # cutoff
                    boxes = cutoff_box(fragment.uuid)
                else:
                    # background
                    boxes = bg_box(fragment.uuid)
                frag_name = term.term_node['Name']
            else:
                # subfragment
                top_frag = term.term_node.top()
                subfrags.append(top_frag)
                boxes = subfrag_box(fragment.uuid, top_frag)
                if term.term_node is top_frag:
                    subfrag_scale = 1.0 / top_frag.exchange_value(scenario, observed=True)
                else:
                    ffs = [f for f in top_frag.io_flows(scenario, observed=True) if f.fragment is term.term_node]
                    if len(ffs) != 1:
                        raise TraversalError('top:%s\nterm:%s\n%d matching flow found' % (top_frag, term.term_node,
                                                                                          len(ffs)))
                    subfrag_scale = 1.0 / ffs[0].node_weight
                # except ZeroDivisionError:
                    # zero exchange value usually means unobserved-- for reference flow that is not usually important
                    # subfrag_scale = 0.0
                print('subfrag scaling: %g' % subfrag_scale)
                frag_name = term.term_node['Name']
                '''
                exchs = [x for x in fragment.get_fragment_inventory(scenario=scenario, scale=node_weight)]
                if len(exchs) > 0:
                    frag_name += '\\\\\n{\scriptsize\n'
                    for x in exchs:
                        frag_name += '%6s: %6.3g %s\\\\' % (x.direction, x.value, x.flow['Name'])
                    frag_name += '}'
                '''
        elif term.is_null:
            # I/O
            boxes = io_box(fragment.uuid)
            frag_name = '%s: %s' % (fragment.direction, fragment.flow['Name'])
        elif term.is_context:
            boxes = context_box(fragment.uuid)
            frag_name = term.term_node.name

        else:
            # process
            boxes = process_box(fragment.uuid)
            frag_name = '[%s] %s' % (term.term_node.origin, term.term_node.name)
            subfrag_scale = term.node_weight_multiplier
        mag_mod = ''
        if fragment.is_balance:
            mag_mod += '{\\sffamily (=)}'
        if node_weight < 0:
            frag_name = '(AVOIDED) ' + frag_name
            mag_mod += '\\darkred'

        if self.comments:
            # this can get added in once we have an easy way to curate comments
            if not first:
                if len(fragment['Comment']) > 0:
                    frag_name += '~$\cdot$~{\\scriptsize %s}' % fragment['Comment']

        boxes += '\n\\rput[l]([angle=0,nodesep=6pt]nx%.5s)' % fragment.uuid
        boxes += '{\parbox{%fcm}{\\footnotesize\\raggedright %s}}' % (parbox_width,
                                                                      tex_sanitize(frag_name))

        if fragment.direction == 'Input':
            arrows = '\\ncline{->}{nx%.5s}{px%.5s}' % (fragment.uuid, fragment.uuid)
        else:
            arrows = '\\ncline{<-}{nx%.5s}{px%.5s}' % (fragment.uuid, fragment.uuid)
        if not first:
            arrows += '\n\\bput(1.18){\\parbox{2cm}{\\centering %s \\scriptsize %.3g %s}}' % (mag_mod,
                                                                                              node_weight,
                                                                                              fragment.flow.unit)

        children = [c for c in fragment.child_flows]
        if 1:  # node_weight != 0 or scenario is None:
            if len(children) > 0 and not term.is_null and not term.is_context:
                parbox_width -= 2.4

                for c in children:
                    arrows += '\n\\ncangle[angleA=-90,angleB=180,framearc=0]{nx%.5s}{px%.5s}' % (fragment.uuid,
                                                                                                 c.uuid)
                    arrows += '\n'
                    boxes += '\n'
                    bplus, aplus, sfplus = self.frag_layout_traverse(c, node_weight * subfrag_scale,
                                                                     scenario=scenario, parbox_width=parbox_width)

                    boxes += bplus
                    arrows += aplus

                    subfrags.extend(sfplus)

        arrows += '\n'
        boxes += '\n'
        return boxes, arrows, subfrags

    def _contrib_chart(self, f_base, results, stages=None, **kwargs):
        """
        Creates a chart and saves it in the image directory. The file name is frag.uuid (the only use of
        the frag argument in the function)
        :param frag: fragment to which the chart pertains
        :param results: a LIST of LciaResult instances (if LciaResults dict, it gets converted to a list)
        :param stages: to be queried
        :return:
        """
        if not isinstance(results, list):
            results = results.to_list()

        if stages is None:
            stages = grab_stages(results)

        fig = scenario_compare_figure(results, stages, center_scale=True, **kwargs)
        save_plot(os.path.join(self.img_folder, self._img_fname(f_base)))
        if stages:
            self._write_stage_names(f_base, stages)
        return fig

    def frag_chart_narrow(self, f_base):
        return '''
\\includegraphics[width=\\textwidth]{%s}
\\rput[tl](pLegend){\\parbox{5cm}{\\raggedright Stages\\\\
{\\scriptsize \\input{%s}}}}
''' % (self.img_rel_path(f_base), self.stg_rel_path(f_base))

    @staticmethod
    def _results_start(title='Contribution Analysis'):
        if title is None:
            return '''
\\begin{minipage}{\\textwidth}
{\\pnode(12,-1){pLegend}}
'''
        else:
            return '''
\\begin{minipage}{\\textwidth}
{\\pnode(12,-1){pLegend}
\\large %s}
''' % title

    @staticmethod
    def _results_end():
        return '\n\\end{minipage}\n'

    def _frag_chart(self, f_base):
        """
        Generates TeX for a contribution analysis chart using the canonical fragment name
        :param f_base:
        :return:
        """
        return '\n\n\\includegraphics[width=\\textwidth]{%s}' % self.img_rel_path(f_base)

    def _scenario_contrib_tables(self, results, stages, scenarios):
        """
        Generates a set of tabular scenario-comparison tables, one for each indicator.
        :param results:
        :param stages:
        :param scenarios:
        :return:
        """
        num_charts = len(results[0])
        chart = '\\\\[1em]\n'

        for i in range(num_charts):
            chart += '\n%s\n\n' % results[0][i].quantity['Name']
            chart += '\n\\begin{tabularx}{\\textwidth}{|X|%s}\n\\hline' % ('r|' * len(results))
            chart += '\\rule[-4pt]{0pt}{16pt}\\textbf{Stage} '
            for j in scenarios:
                chart += ' & \\textbf{%.15s} ' % j

            # chart += '%s \n ' % TAB_LF
            # for j in results:
            #     chart += ' &  %s ' % j.quantity.unit()

            chart += '%s \\hline\n' % TAB_LF

            results_inner = [res[i] for res in results]

            chart += self._contrib_table_body(results_inner, stages=stages)
            chart += '%s\n \\hline\\end{tabularx}\n' % TAB_LF

        return chart

    @staticmethod
    def _contrib_table_body(results_inner, stages=None):
        if stages is None:
            stages = grab_stages(results_inner)

        chart = ''
        for i, s in enumerate(stages):
            chart += '%s -- %s' % (chr(ord('A') + i), tex_sanitize(s))

            for r in results_inner:
                d = sum([q for q in r.contrib_query([s])])
                if abs(d) > 0.1 * r.span:
                    chart += ' & \\sffamily \\textbf{%8.2e} ' % d
                elif d == 0:
                    chart += ' & -- '
                else:
                    chart += ' & \\sffamily %8.2e ' % d

            chart += '%s\n' % TAB_LF

        chart += '\\hline \\rule[-3pt]{0pt}{12pt} \\hfill \\textbf{TOTAL:} '

        for r in results_inner:
            chart += ' & \sffamily  \\textbf{%6.3e} ' % r.total()

        return chart

    def _contrib_table(self, results, stages=None):
        """
        Generates TeX for a Tabular contribution chart.
        :param results:
        :param stages:
        :return:
        """
        if not isinstance(results, list):
            results = results.to_list()

        chart = '\n\\begin{tabularx}{\\textwidth}{|X|%s}\n\\hline' % ('r|' * len(results))
        chart += '\\rule[-4pt]{0pt}{16pt}\\textbf{Stage} '
        for i, j in enumerate(results):
            chart += ' & \\textbf{%.12s} ' % j.quantity['Indicator']

        chart += '%s \n ' % TAB_LF
        for j in results:
            chart += ' &  %s ' % j.quantity.unit

        chart += '%s \\hline\n' % TAB_LF

        chart += self._contrib_table_body(results, stages=stages)
        chart += '%s\n \\hline\\end{tabularx}\n' % TAB_LF

        return chart

    def _process_tex(self, frag, scenario=None, full=True):
        """
        Returns TeX content for the process tree figure
        :param frag:
        :param scenario:
        :param full: [True] whether to print section heading and inventory
        :return:
        """
        coords = self.frag_layout_recurse(frag, scenario=scenario)

        if full:
            tex_dump = fragment_header(frag)
            tex_dump += fragment_fu(frag, scenario=scenario)
            tex_dump += fragment_inventory(frag, scenario=scenario)
        else:
            tex_dump = fragment_fu(frag, scenario=scenario)

        tex_dump += frag_drawing_opener(coords[-1][2])
        tex_dump += frag_pnodes(coords)
        bx, ar, subfrags = self.frag_traversal_entry(frag, scenario=scenario)  # subfrags go back up to the author
        tex_dump += bx
        tex_dump += ar
        tex_dump += frag_drawing_closer()

        return tex_dump, subfrags

    def results_tex(self, f_base, results, stages=None, table=False, scenarios=None, full=True, **kwargs):
        """
        Generates a chart and table from a set of results, and returns the TeX content
        :param f_base:
        :param results:
        :param stages:
        :param table:
        :param scenarios: if results is a double-array, indicate scenario names
        :param full: [True] if false, suppress heading
        :param kwargs:
        :return:
        """
        if not isinstance(results, list):
            results = results.to_list()
        if stages is None:
            stages = grab_stages(results)

        skip = False
        if scenarios is None:
            title = 'Contribution Analysis'
            if all(r.total() == 0 for r in results):
                skip = True
        else:
            title = 'Scenario Analysis'
            if all(r.total() == 0 for k in results for r in k):
                skip = True
        if not full:
            title = None

        tex_dump = self._results_start(title=title)

        if skip:
            tex_dump += '\n\nNo Impacts.\n'

        else:
            try:
                self._contrib_chart(f_base, results, stages=stages, scenarios=scenarios, **kwargs)
                tex_dump += self._frag_chart(f_base)
            except AttributeError:
                print('bailing out of chart')
                tex_dump = ''

            if table is True:
                if scenarios is None:
                    tex_dump += self._contrib_table(results, stages=stages)
                else:
                    tex_dump += self._scenario_contrib_tables(results, stages, scenarios)

        tex_dump += self._results_end()
        return tex_dump

    def results_report(self, f_base, results, **kwargs):
        """

        :param f_base:
        :param results:
        :param kwargs: stages, table, other args to scenario_compare_figure
        :return:
        """
        self._write_output(f_base, self.results_tex(f_base, results, **kwargs), add_to_index='doc')

    def results_float(self, f_base, results, **kwargs):
        """
        Just like results_report, but suppress the large heading

        :param f_base:
        :param results:
        :param kwargs: stages, table, other args to scenario_compare_figure
        :return:
        """
        self._write_output(f_base, self.results_tex(f_base, results, full=False, **kwargs), add_to_index='doc')

    def fragment_diagram(self, f_base, frag, scenario=None):
        """
        Generate just the fragment diagram, without heading, inventory or results
        :param f_base:
        :param frag:
        :param scenario:
        :return:
        """
        tex_dump, subfrags = self._process_tex(frag, scenario=scenario, full=False)
        self._write_output(f_base, tex_dump, add_to_index='doc')
        return subfrags

    def fragment_report(self, frag, f_base=None, scenario=None, stages=None, results=None, table=False, **kwargs):
        """
        Generate a report for the supplied fragment in the TeXAuthor working directory.  Fragment report is
        created as a subsection with a heading, hyperlink, and full inventory listing.
        If results arg is present, draw a chart of stage contributions.
        if table is True, draw a table of stage contributions.
        :param frag:
        :param f_base: defaults to frag.uuid; if the default is overridden, output is written to doc
        :param scenario:
        :param stages: list of stages to report (defaults to a collection of all stages found in the results
        :param results: list of LciaResult objects
        :param table: [False]
        :return: a list of the fragment's children
        """
        if f_base is None:
            f_base = frag.uuid
            _to_index = 'tex'
        else:
            _to_index = 'doc'

        tex_dump, subfrags = self._process_tex(frag, scenario=scenario)

        if results is not None:
            tex_dump += self.results_tex(f_base, results, stages=stages, table=table, **kwargs)

        self._write_output(f_base, tex_dump, add_to_index=_to_index)

        if frag.has_property('ModelDocumentation'):
            docbase = os.path.join(self.folder, '%s-model-doc' % f_base)
            self._write_output(docbase, frag['ModelDocumentation'], add_to_index='doc')

        return subfrags

    def new_report(self):
        """
        Create new fragment-data and model-doc files; if fragment
        :return:
        """
        self._file_list = []
        self._init_report_file(self._wrapper_fname,
                               'TeX report wrapper for fragment-data, lcatools')

        self._init_report_file(self._documentation_fname,
                               'TeX report wrapper for model documentation, lcatools')

        self._seen = set()

    def new_section(self, section_name):
        with open(self._wrapper_fname, 'a') as fp:
            fp.write('\\cleardoublepage\n\n\\section{%s}\n\n' % section_name)

    ''' # we don't have those kinds of queries anymore
    def report(self, query, chart=True):
        """
        generate TeX reports for the fragments and results.
        :param query: a ForegroundQuery
        :param chart: [True] whether to compute + plot results
        :return: list of child fragments
        """
        children = []
        for i, f in enumerate(query.fragments):
            if chart:
                ch = self.fragment_report(f, stages=query.agg_stages[i], results=query.agg_results[i], table=True,
                                          scenario=query.scenario)
            else:
                ch = self.fragment_report(f, scenario=query.scenario)
            for k in ch:
                if k not in children:
                    # preserve order encountered
                    children.append(k)

        return children
    '''

    def run_frag(self, f, scenario, quantities, **kwargs):
        res = [f.fragment_lcia(k, scenario=scenario).aggregate() for k in quantities]
        self._seen.add(f)
        children = self.fragment_report(f, scenario=scenario, results=res, **kwargs)
        new_children = [ch for ch in children if ch not in self._seen]
        return new_children

    def recurse_report(self, *fragments, quantities=None, section_names=None, scenario=None,
                       single_origin=True, **kwargs):
        """
        Recursively generate fragment drawings for supplied fragments plus all descendants.
        TODO: figure out a way to organize / group child fragments other than by tier.
        Current plans are to _manually edit_ fragment_data.tex to reorganize sections / hide private fragments
        :param fragments: positional arguments are fragments at the top-level
        :param quantities: optional list of quantities for LCIA (3 is about the practical limit)
        :param section_names: [None] a list of names for successive tiers
        :param scenario:
        :param single_origin: restrict recursion to fragments in the same foreground
        :param kwargs: stages [None] - list of stages (default is to autodetect)
                       table [False] - draw table
                        full [True] - if False, omit title
                        **kwargs - passed to scenario_compare_figure
        """
        self.new_report()
        if section_names is None:
            section_names = []

        try:
            section_name = section_names.pop(0)
        except IndexError:
            section_name = 'Top-level Fragments'

        # run out top-level frags, then draw their children in the next section
        if quantities is None:
            quantities = ()

        level = 0

        while len(fragments) > 0:
            if len(section_name) > 0:
                self.new_section(section_name)
            #query = cat.fg_query(fragments, qs, **kwargs)
            #children = self.report(query)
            next_children = []
            for f in fragments:
                org = f.origin
                new_children = self.run_frag(f, scenario, quantities, **kwargs)
                if single_origin:
                    new_children = [ch for ch in new_children if ch.origin == org]
                next_children.extend(new_children)
            fragments = next_children

            level += 1

            try:
                section_name = section_names.pop(0)
            except IndexError:
                section_name = 'Tier %d Fragments' % level



import matplotlib.pyplot as plt
import numpy as np
from matplotlib.cm import get_cmap


def sensitivity_bars(df, baseline='Baseline', lo='Low_5', hi='High_95', colors=None, cmap='tab10'):
    """
    Requires a dataframe with columns "Case", "Variant", "Quantity", "Indicator", and baseline, lo, and hi columns.

    Creates a grid where each row is a quantity, each column is a case, and each facet is a barchart where individual
    bars are variants.  Draws a bar of height baseline and errorbars spanning from lo to hi.

    :param df:
    :param baseline: 'Baseline' Column containing baseline data
    :param lo: 'Low_5' Column containing low-sens data
    :param hi: 'High_95' Column containing high-sens data
    :param colors: None a list of colors corresponding to variants
    :param cmap: 'tab10' if no colors are specified, the colormap from which to draw colors
    :return:
    """
    """
    :param df: 
    :return: 
    """
    # Get unique cases and quantities
    cases = df['Case'].unique()
    quantities = df['Quantity'].unique()
    variants = df['Variant'].unique()

    # Create subplot grid
    fig, axes = plt.subplots(
        len(quantities), len(cases), figsize=(2 * len(cases), 1.6 * len(quantities)), sharey=False, sharex=True
    )

    # Ensure axes is a 2D array for consistent indexing
    if len(quantities) == 1:
        axes = [axes]
    if len(cases) == 1:
        axes = np.array(axes).reshape(-1, 1)

    # Define colors for variants
    n = len(variants)
    if colors is None:
        colors = get_cmap(cmap, n)(np.arange(len(variants)))

    # store ylims by quantity
    row_y_limits = []
    # Iterate through cases and quantities to plot
    for i, quantity in enumerate(quantities):
        y_min, y_max = float('inf'), float('-inf')
        for j, case in enumerate(cases):
            ax = axes[i][j]

            # Filter data for the current case and quantity
            sub_df = df[(df['Case'] == case) & (df['Quantity'] == quantity)]

            # X positions for the bars
            x_positions = np.arange(len(variants))

            # Plot bars with error bars
            ax.bar(
                x_positions,
                sub_df[baseline],
                yerr=[
                    sub_df[baseline] - sub_df[lo],
                    sub_df[hi] - sub_df[baseline]
                ],
                width=0.66,
                capsize=4,
                color=colors[:len(sub_df)],
                edgecolor='none',
                tick_label=sub_df['Variant']
            )

            # Set subplot title and labels
            if i == 0:
                ax.set_title(f"Case: {case}")
            if j == 0:
                ax.set_ylabel(sub_df.iloc[0]['Indicator'])

            ax.set_xticks(x_positions)
            ax.set_yticks([])
            ax.set_xlim([x_positions[0] - 0.7, x_positions[-1] + 0.7])
            ax.set_xticklabels(sub_df['Variant'], rotation=45, ha='right')
            # Update y-axis limits
            current_min, current_max = ax.get_ylim()
            y_min = min(y_min, current_min)
            y_max = max(y_max, current_max)

        # Store the calculated limits for the row
        row_y_limits.append((y_min, y_max))
        for j in range(len(cases)):
            axes[i][j].set_ylim((y_min, y_max))

    return fig



from antelope_core.entities.xlsx_editor import XlsxUpdater
from antelope_foreground.entities.fragments import InvalidParentChild
from antelope import EntityNotFound, UnknownOrigin


from pandas import to_numeric


def prepare_dataframe(df):
    return df.apply(to_numeric).sort_values(df.columns[0])


def write_results_dataframe_to_gsheet(gsheet, sheetname, df, clear_sheet=True, write_header=True, header_levels=None,
                                      fillna='NA', write_index=True):
    """

    :param gsheet: a GoogleSheetReader
    :param sheetname: sheet to write to or create
    :param df: sorted, prepared data
    :param clear_sheet: [True]
    :param write_header: [True] whether to write header (False: leave it standing)
    :param header_levels: number of header levels to write. Must be <= nlevels
    :param fillna:
    :param write_index:
    :return:
    """

    ncol = len(df.columns)
    if header_levels is None or header_levels > df.columns.nlevels:
        header_levels = df.columns.nlevels

    if sheetname in gsheet.sheet_names():
        # start by clearing the sheet- with or without headers
        if clear_sheet:
            if write_header:
                gsheet.clear_region(sheetname)
            else:
                gsheet.clear_region(sheetname, start_row=header_levels)
        else:
            if write_header:
                gsheet.clear_region(sheetname, end_col=ncol, end_row=header_levels-1)
    else:
        gsheet.create_sheet(sheetname)

    #then populate
    def _row_gen(_df):
        for _i, row in _df.fillna(fillna).iterrows():
            if write_index:
                yield [_i] + list(row.values)
            else:
                yield list(row.values)

    if write_header:
        for i in range(header_levels):
            h = [''] + list(df.columns.get_level_values(i))
            gsheet.write_row(sheetname, i, h)
    gsheet.write_rectangle_by_rows(sheetname, _row_gen(df), start_row=header_levels)



class XlsxForegroundUpdater(XlsxUpdater):
    """
    This subclass introduces the ability to read a ModelFlows sheet for both flows and fragments.  A "flows" sheet
    can also be read if present, but is not required.

    The worksheet with the model flows must have the following mandatory (*) and optional (-) columns (case-sensitive)
    (note that these fields do not have to be defined for every fragment, just present as columns):

    [for flow definition]
     * 'flow' or 'flow_ref' -- external_ref for flow
     * 'ref_quantity' -- external ref for flow's reference quantity
     - 'flowname' -- name of flow (if missing, will be created from the external_ref

    [for fragment definition]
     * 'external_ref' -- name of fragment (blank = unobserved fragment)
     * 'parent' -- external_ref of parent fragment
     * 'direction' -- along with flow above, direction of flow [human-sensible-- see note]
     * 'balance' -- mark 'True' if the fragment is a balancing fragment
     * 'Name' -- fragment name
     * 'StageName' -- fragment stage
     - 'Comment' -- optional

    [for fragment termination]
     * 'exchange_value' -- observed exchange value
     - 'units' -- for conversion -- must be known units to the quantity named in ref_quantity
     * 'termination' --
     * 'term_flow'
     * 'descend'
    """
    @property
    def flow_key(self):
        return self._flow_key

    @flow_key.setter
    def flow_key(self, value):
        if self._flow_key is None:
            self._flow_key = value
        elif value == self._flow_key:
            return
        else:
            raise ValueError('Flow Key mismatch! %s vs %s' % (value, self.flow_key))

    def _populate_frags(self, model_sheet):
        sh, headers = self._sheet_accessor(model_sheet)
        if sh is None:
            return []

        if 'flow' in headers:
            self.flow_key = 'flow'
        elif 'flow_ref' in headers:
            self.flow_key = 'flow_ref'
        else:
            raise AttributeError('Model sheet %s missing flow definition' % model_sheet)

        frags = [{headers[i]: self._grab_value(k) for i, k in enumerate(sh.row(row))} for row in range(1, sh.nrows)]
        return list(filter(lambda x: x.get(self.flow_key), frags))  # filter out entries with blank flow

    def __init__(self, fg, xlrd_like, *model_sheets, **kwargs):
        super(XlsxForegroundUpdater, self).__init__(xlrd_like, **kwargs)
        self._fg = fg
        self._flow_key = None

        self._model_data = []
        if len(model_sheets) == 0:
            model_sheets = ['ModelFlows']

        for model_sheet in model_sheets:
            self._model_data.extend(self._populate_frags(model_sheet))
        self._unrec = []

    @property
    def ar(self):
        return self._fg

    @property
    def origin(self):
        return self._fg.origin

    @property
    def qi(self):
        return self._fg

    _ext_ref_sub = str.maketrans('_-', '  ')

    def _ext_ref_name(self, ext_ref):
        return ext_ref.translate(self._ext_ref_sub).title()

    def _new_entity(self, etype, rowdata):
        if etype == 'flow':
            ext_ref = rowdata.pop('external_ref')
            ref_q = rowdata.pop('referenceQuantity')
            name = rowdata.pop('Name', None) or self._ext_ref_name(ext_ref)
            e = self._fg.add_or_retrieve(ext_ref, ref_q, name, **rowdata)
            if e.entity_type != 'flow':
                raise TypeError('Type conflict encountered on %s [%s != %s]' % (ext_ref, etype, e.entity_type))
        elif etype == 'quantity':
            ext_ref = rowdata.pop('external_ref')
            ref_q = rowdata.pop('referenceUnit')
            name = rowdata.pop('Name', None) or self._ext_ref_name(ext_ref)
            e = self._fg.add_or_retrieve(ext_ref, ref_q, name, **rowdata)
            if e.entity_type != 'quantity':
                raise TypeError('Type conflict encountered on %s [%s != %s]' % (ext_ref, etype, e.entity_type))
        elif etype == 'model_flow':
            ext_ref = rowdata[self.flow_key]
            ref_q = rowdata['ref_quantity']
            if ext_ref.find('/') > 0:
                origin, external_ref = ext_ref.split('/', maxsplit=1)
                name = rowdata.get('flowname') or self._ext_ref_name(external_ref)
                e = self._fg.add_or_retrieve(external_ref, ref_q, name, origin=origin)
            else:
                name = rowdata.get('flowname') or self._ext_ref_name(ext_ref)
                e = self._fg.add_or_retrieve(ext_ref, ref_q, name)
            if e.entity_type != 'flow':
                raise TypeError('Type conflict encountered on %s [%s != %s]' % (ext_ref, 'flow', e.entity_type))
        else:
            raise KeyError('Unknown entity type %s' % etype)

    @property
    def frags(self):
        return self._model_data

    def _grab_model_flows(self):
        """
        First pass through model flows: grab and create flow names so that they can be assigned flow properties
        :return:
        """
        for frag in self.frags:
            try:
                self.get_flow(frag[self.flow_key])
            except EntityNotFound:
                self._new_entity('model_flow', frag)

    def _find_frag(self, frag):
        if frag['external_ref'] is None:
            # infer frag from parent and flow
            parent = self.ar[frag['parent']]
            flow = self.get_flow(frag[self.flow_key])
            try:
                return next(parent.children_with_flow(flow))
            except StopIteration:
                return None
        else:
            return self.ar[frag['external_ref']]  # could itself be None

    def _grab_model_frags(self):
        for frag in self.frags:
            bal = bool(frag.get('balance'))
            if frag['parent'] is None:
                parent = None
            else:
                parent = self.ar[frag['parent']]
                # parent.to_foreground()  # this is now accomplished in fragment constructor via set_parent()
            f = self._find_frag(frag)
            if f is None:
                f = self.ar.new_fragment(frag[self.flow_key], frag['direction'], parent=parent,
                                         balance=bal, Name=frag.get('Name'), StageName=frag.get('StageName'),
                                         Comment=frag.get('Comment'), external_ref=frag['external_ref'])
                if f['Name'] == f.uuid:
                    f['Name'] = f.flow['Name']  # just for human readability

            else:
                if f.flow.external_ref != frag[self.flow_key]:
                    f.flow = self.get_flow(frag[self.flow_key])
                if f.reference_entity != parent:
                    f.unset_parent()
                    f.set_parent(parent)
                if bal != f.is_balance:
                    if bal:
                        try:
                            f.set_balance_flow()
                        except InvalidParentChild:
                            print('Reference Fragment %s cannot set balance' % f)
                    else:
                        f.unset_balance_flow()

                for k in ('Name', 'StageName', 'Comment'):
                    s = frag.get(k)
                    if s is None:
                        continue
                    f[k] = s

    def _terminate_model_frags(self):
        for frag in self.frags:
            f = self._find_frag(frag)
            if f is None:
                raise KeyError('Unable to find this frag! %s' % frag)
            if frag.get('exchange_value') is not None:
                self.ar.observe(f, exchange_value=float(frag['exchange_value']), units=frag.get('units'))
            if frag.get('termination') is None:
                if len(list(f.child_flows)) == 0:
                    f.clear_termination()
                # else- nothing to do- just
                continue
            else:
                if frag['termination'] == 'self':
                    f.to_foreground()
                    continue
                try:
                    term = self.ar.get(frag['termination'])
                except (EntityNotFound, UnknownOrigin):
                    self._unrec.append((f.external_ref, frag['termination']))
                    continue
                if f.term.is_null or f.term.term_node != term:
                    f.clear_termination()
                    # if term.entity_type == 'process':
                    #     f.set_background()  # processes are background-terminated ## outmoded
                    tflow = frag.get('term_flow')
                    if tflow is not None:
                        tflow = self.get_flow(tflow)
                    f.terminate(term, term_flow=tflow)
                desc = {'true': True,
                        'false': False,
                        '0': None}[frag.get('descend', '0').lower()]  # map to bool
                if desc is None:
                    continue
                f.term.descend = desc

    def apply(self):
        for etype in ('quantity', 'flow'):  # these are the only types that are currently handled
            self._process_sheet(etype)
        self._grab_model_flows()
        self._process_flow_properties()
        self._grab_model_frags()
        self._terminate_model_frags()

        if len(self._unrec) > 0:
            print('Unrecognized Terminations: ')
            for frag, term in self._unrec:
                print('  [%s] -> %s' % (frag, term))

    def get_flow(self, flow):
        return self.ar.get(flow)

    def get_context(self, cx):
        return self.ar.context(cx)

    '''
    parameters
    read + update sheet
    '''
    def read_parameters(self, update_default=False):
        """
        Apply all read parameters to the model.
        :param update_default: whether to reset default (observed) values as well
        :return:
        """
        sh, headers = self._sheet_accessor('parameters')
        if sh is None:
            return

        for row in range(1, sh.nrows):
            rowdata = {headers[i]: self._grab_value(k) for i, k in enumerate(sh.row(row))}
            units = rowdata.pop('flow_unit', None)
            knob = self._fg[rowdata.pop('parameter', None)]
            if knob:
                if update_default:
                    self._fg.observe(knob, exchange_value=rowdata.pop('default'), units=units)







"""
A collection of handy functions for easily displaying information
"""


def process_screen(runner, _ind):
    """
    Show the most important processes. Right now just prints them all out.
    :param runner: A ScenarioRunner
    :param _ind:
    :return:
    """
    for s in runner.scenarios:
        r = runner.result(s, _ind)
        if r.total() == 0:
            continue
        print('\n%s' % s)
        r.terminal_nodes().show_components()


def substance_screen(runner, indicator, use_abs=True):
    """
    Put in a scenario runner and an indicator; generates a list of the most important compounds in each scenario
    (based on a flattened lcia result)
    :param runner: scenario runner
    :param indicator: indicator
    :param use_abs: [True] whether to strictly use absolute values to evaluate significance (the default is True)
    :return:
    """
    for s in runner.scenarios:
        r = runner.result(s, indicator).flatten()
        if use_abs:
            tot = sum(abs(c.cumulative_result) for c in r.components())
            thresh = 0.0045
        else:
            tot = r.total()
            thresh = 0.009
        if tot == 0:
            continue
        print('\n%s' % s)
        print(r)
        ems = sorted(r.components(), key=lambda x: abs(x.cumulative_result), reverse=True)
        c = d = 0
        _pf = True
        for em in ems:
            if use_abs:
                delta = abs(em.cumulative_result)
            else:
                delta = em.cumulative_result
            if _pf:
                c += delta
                print('%2.6g %% %s' % (100*delta/tot, em))
                if abs(delta) / abs(tot) < thresh:
                    _pf = False
            else:
                d += delta
        print('%2.6g %% remainder' % (100 * d / tot))


def show_top_n(res, threshold=0.0075):
    """
    An abortive attempt to to substance_screen for a single result.
    This should probably be an LciaResult method
    :param res:
    :param threshold:
    :return:
    """

    print(res)
    _ins = []
    _edz = []
    _thr = abs(res.total() * threshold)
    for dd in sorted(res.components(), key=lambda _x: _x.cumulative_result):
        if abs(dd.cumulative_result) > _thr:
            print(dd)
            _ins.append(dd)
        else:
            _edz.append(dd)
    _s = sum(_z.cumulative_result for _z in _edz)
    print('%d entries, %d saved' % (len(_ins + _edz), len(_ins)))
    print(': %g : Remainder = %2.2f %%' % (_s, 100*_s/res.total()))
    # return _ins


def nodes(_frag):
    """
    Print a list of terminal nodes contained within the fragment and subfragments.
    Should probably be an LcFragment method.
    :param _frag:
    :return:
    """
    _ns = set()
    for sc, t in _frag.terminations():
        if t.is_process:
            _ns.add(t.term_node)
        elif t.is_subfrag:
            _ns |= nodes(t.term_node)
    for cf in _frag.child_flows:
        if cf.term.is_context:
            _ns.add(_frag)
        else:
            _ns |= nodes(cf)
    return _ns





from .scenario_runner import ScenarioRunner
from antelope_foreground.fragment_flows import group_ios, ios_exchanges, frag_flow_lcia

from collections import defaultdict


class SensitivityRunner(ScenarioRunner):
    @classmethod
    def run_lca(cls, model, qs, *common, agg_key=None, sens_hi=None, sens_lo=None,
                hi_sense=None, lo_sense=None, **scenarios):
        """
        "Do everything" classmethod for running an LCA study based on a unitary model.
        Prior to calling this: all scenarios must be properly prepared (parameters and terminations observed)

        WHEN calling this, the two required positional arguments are: the model (fragment) and an iterable of LCIA
        methods (quantities or quantity refs).

        Optional positional parameters are common scenarios to be added to all cases

        sens_hi and sens_lo (synonyms: hi_sense and lo_sense) are optional tuples of scenario specifications for
        high- and low-sensitivity test cases, respectively.  These will not get run if they are omitted.

        Finally, the scenarios themselves are supplied as kwarg: tuple. At least one must be supplied, even if it
        has None value, because otherwise the runner won't know what to name the default scenario.
        :param model:
        :param qs:
        :param common:
        :param agg_key:
        :param sens_hi:
        :param sens_lo:
        :param hi_sense: synonym for sens_hi
        :param lo_sense: synonym for sens_lo
        :param scenarios:
        :return:
        """
        run = cls(model, *common, agg_key=agg_key, sens_hi=sens_hi or hi_sense, sens_lo=sens_lo or lo_sense)
        for k, v in scenarios.items():
            run.add_case(k, v)

        for q in qs:
            run.run_lcia(q)

        return run

    def __init__(self, model, *common_scenarios, sens_hi=None, sens_lo=None, **kwargs):
        super(SensitivityRunner, self).__init__(model, *common_scenarios, **kwargs)

        self._results_hi = dict()
        self._results_lo = dict()

        self._traversals_hi = dict()
        self._traversals_lo = dict()

        self._sens_hi = self._scenario_tuple(sens_hi)
        self._sens_lo = self._scenario_tuple(sens_lo)

    def add_hi_sense(self, param):
        self._sens_hi += self._scenario_tuple(param)
        for case in self.scenarios:
            self._traverse_hi(case)

    def add_lo_sense(self, param):
        self._sens_lo += self._scenario_tuple(param)
        for case in self.scenarios:
            self._traverse_lo(case)

    def _traverse_hi(self, case):
        sc = self._params[case]
        sc_apply = sc + tuple(self.common_scenarios)

        sc_hi = sc_apply + self._sens_hi
        self._traversals_hi[case] = self._model.traverse(scenario=sc_hi)

    def _traverse_lo(self, case):
        sc = self._params[case]
        sc_apply = sc + tuple(self.common_scenarios)

        sc_lo = sc_apply + self._sens_lo
        self._traversals_lo[case] = self._model.traverse(scenario=sc_lo)

    def _traverse_case(self, case):
        print('traversing %s' % case)
        sc = self._params[case]
        sc_apply = sc + tuple(self.common_scenarios)
        self._traversals[case] = list(self._model.traverse(sc_apply))

        if self._sens_hi:
            self._traverse_hi(case)

        if self._sens_lo:
            self._traverse_lo(case)

    def inventory_hi(self, scenario, **kwargs):
        ios, _ = group_ios(self._model, self._traversals_hi[scenario], **kwargs)
        return ios_exchanges(ios, ref=self._model)

    def inventory_lo(self, scenario, **kwargs):
        ios, _ = group_ios(self._model, self._traversals_lo[scenario], **kwargs)
        return ios_exchanges(ios, ref=self._model)

    def _run_scenario_lcia(self, scenario, lcia, **kwargs):
        sc = self._params[scenario]
        sc_apply = sc + tuple(self.common_scenarios)

        res = frag_flow_lcia(self._traversals[scenario], lcia, scenario=sc_apply, **kwargs)

        if self._sens_hi:
            sc_hi = sc_apply + self._sens_hi
            res_hi = frag_flow_lcia(self._traversals_hi[scenario], lcia, scenario=sc_hi, **kwargs)
        else:
            res_hi = res

        if self._sens_lo:
            sc_lo = sc_apply + self._sens_lo
            res_lo = frag_flow_lcia(self._traversals_lo[scenario], lcia, scenario=sc_lo, **kwargs)
        else:
            res_lo = res

        self._results_lo[scenario, lcia] = min([res, res_lo, res_hi], key=lambda x: x.total())
        self._results_hi[scenario, lcia] = max([res, res_lo, res_hi], key=lambda x: x.total())

        return res

    sens_order = ('result', 'result_lo', 'result_hi')

    def sens_result(self, scenario, lcia_method):
        return (self._results[scenario, lcia_method],
                self._results_lo[scenario, lcia_method],
                self._results_hi[scenario, lcia_method])

    @property
    def results_headings(self):
        return ['scenario', 'stage', 'alt_stage', 'method', 'category', 'indicator', 'result', 'result_lo', 'result_hi', 'units']

    def _gen_lcia_rows(self, scenario, q, include_total=False, aggregate=True, **kwargs):
        """
        This is really complicated because we don't know (or don't want to assume) that the three scores will have
        the same stages-- because low and hi scenarios could trigger different traversals / terminations.

        it certainly makes the code look like hell.
        the code makes an open ended dict of stages, with a subdict of result, result_lo, result_hi
        these get populated only when encountered, and output only when present.
        :param scenario:
        :param q:
        :param include_total:
        :return:
        """
        keys = defaultdict(dict)
        if aggregate:
            ress = [k.aggregate(key=self._agg) for k in self.sens_result(scenario, q)]
            for i, res in enumerate(ress):
                for c in res.components():
                    keys[c.entity][self.sens_order[i]] = c.cumulative_result
        else:
            ress = [k for k in self.sens_result(scenario, q)]
            for i, res in enumerate(ress):
                for c in res.components():
                    stgs = self._agg(c.entity)
                    keys[stgs, c.uuid][self.sens_order[i]] = c.cumulative_result

        for key, result in sorted(keys.items(), key=lambda x: x[0]):
            if aggregate:
                stage, alt_stage = key
            else:
                stage, alt_stage = key[0]  # double packed
            d = {
                'scenario': str(scenario),
                'stage': stage,
                'alt_stage': alt_stage,
                'result': None,
                'result_lo': None,
                'result_hi': None
            }
            for k, v in result.items():
                d[k] = self._format(v)

            yield self._gen_row(q, d)
        if include_total:
            dt = {
                'scenario': str(scenario),
                'stage': 'Net Total'
            }
            for i, k in enumerate(ress):
                dt[self.sens_order[i]] = k.total()

            yield self._gen_row(q, dt)




from antelope_core.contexts import NullContext
from antelope import comp_dir


class AllocationEngine(object):
    """
    Creates an allocation framework for a spanner. By supplying an allocation flow (and a reference quantity), all
    cutoffs that are commensurable to the ref qty are cast into the allocation flow, and are aggregated during
    traversal. This fragment can be driven backwards (by anchoring to the allocation flow) to allocate the
    upstream impacts (and subsequent cut-offs).  This will emanate the spanner's reference flow in a fractional amount
    equal to the allocation share.

    Single-product wrappers then encapsulate that aggregator by terminating to the allocation flow, optionally capping
    off the allocation fraction.
    """
    @classmethod
    def all_flows(cls, fg, spanner, alloc_flow, external_ref=None, scenario=True):
        ae = cls(fg, spanner, alloc_flow, external_ref=external_ref)
        for c in spanner.cutoffs(scenario):
            ae.allocate_flow(c.flow, c.direction)

        return ae

    def __init__(self, fg, spanner, alloc_flow, external_ref=None):
        """
        We store the spanner
        :param fg: the foreground to contain the constructed fragments
        :param spanner: the multi-output model to be allocated
        :param alloc_flow: the flow (and reference quantity) to accumulate the allocated output
        :param external_ref: [None] default: 'allocator_[spanner.external_ref]_[alloc_flow.external_ref]'
        """
        self._fg = fg
        self._spanner = spanner
        self._alloc_flow = alloc_flow
        if external_ref is None:
            external_ref = 'allocator_%s_%s' % (self._spanner.external_ref, alloc_flow.external_ref)

        self._allocator = external_ref
        self._build_allocation_container()

    @property
    def alloc_flow(self):
        return self._alloc_flow

    @property
    def allocator(self):
        return self._fg[self._allocator]

    def allocate_flow(self, flow, direction='Output'):
        """
        Adds a child flow to the allocator, and attaches a balance flow converting it to the allocation flow. If the
        flow is not generated by the embedded model, it will not be driven and will show zero activity.

        Note that the direction is nominal- a subfragment cannot emanate the same flow in two directions because of
        subfragment aggregation, and the balance flow will enforce the proper magnitude regardless of direction.
        Fragment allocation will also report the aggregated flow with the correct natural direction.   Provide
        direction='Input' if your child flow is an input and you want to avoid negative-signed outputs within
        the internal traversal.
        :param flow:
        :param direction: default 'Output'
        :return:
        """
        alloc_qty = self.alloc_flow.reference_entity
        container = self.allocator

        if flow.cf(alloc_qty) != 0:
            try:
                j = next(container.children_with_flow(flow))
            except StopIteration:
                j = self._fg.new_fragment(flow, direction, parent=container)
                self._fg.new_fragment(self.alloc_flow, direction, parent=j, balance=True)

            j.balance_flow.flow = self.alloc_flow

    def _build_allocation_container(self):
        """
        Builds an Allocation fragment or "allocator": anchors the target spanner inside a superfragment, then catches
        all emerging allocatable outputs and casts them to an allocation flow.

        encapsulates the target activity inside a node, and then converts all
        allocatable outputs of the activity to their allocation quantity.
        - take the activity (spanner) to allocate
        - create a new fragment with the same reference flow
        - anchor it to the target spanner

        :return:
        """
        alloc_qty = self.alloc_flow.reference_entity

        if self.allocator is None:
            container = self._fg.new_fragment(flow=self._spanner.flow, direction=comp_dir(self._spanner.direction),
                                              exchange_value=self._spanner.cached_ev,
                                              external_ref=self._allocator)
        else:
            container = self.allocator

        self._fg.observe(container, exchange_value=self._spanner.observed_ev)
        container['alloc_flow'] = self.alloc_flow.link
        container.terminate(self._spanner)

        return container

    def build_allocated_product(self, product_flow, direction='Output', external_ref=None, cap_activity=True):
        """
        This builds a single-output allocated production process for the named flow, using the previously
        constructed allocator.
        This caps off the facility activity flow and delivers a desired product flow with allocated burdens

        :param product_flow:
        :param direction: of product w.r.t allocated activity [default 'Output']
        :param external_ref: default: '[allocator.external_ref]_alloc_[product_flow.external_ref]
        :param cap_activity: [True] conceal the allocator's activity share
        :return:
        """
        if external_ref is None:
            external_ref = '%s_alloc_%s' % (self.allocator.external_ref, product_flow.external_ref)
        if self._fg[external_ref] is None:
            capsule = self._fg.new_fragment(flow=product_flow, direction=direction, exchange_value=1.0,
                                            external_ref=external_ref)
        else:
            capsule = self._fg[external_ref]
        self._fg.observe(capsule)
        if capsule.balance_flow is None:
            j = self._fg.new_fragment(flow=self.alloc_flow, direction=comp_dir(direction), parent=capsule, balance=True)
        else:
            j = capsule.balance_flow
            j.flow = self.alloc_flow
        j.terminate(self.allocator, term_flow=self.alloc_flow)

        if cap_activity:
            try:
                k = next(j.children_with_flow(self.allocator.flow))
            except StopIteration:
                k = self._fg.new_fragment(self.allocator.flow, direction=comp_dir(self.allocator.direction), parent=j)
            k.terminate(NullContext)
        else:
            try:
                k = next(j.children_with_flow(self.allocator.flow))
                self._fg.delete(k)
            except StopIteration:
                pass

        return capsule



"""
Combination charts.
 - Horizontal aspect;
 - Positive + Negative bars + net bar;
 - multi-impact;
 - scenario comparisons.

These are expert charts. Nonexpert users have reported their "brain entirely shuts down" when presented with
more elaborate instances of these charts.
"""

import matplotlib.pyplot as plt
from .base import color_range, label_segment, has_pos_neg, has_nonzero, save_plot, hue_from_string


def _one_bar(ax, pos_y, neg_y, data, hue, units, threshold, legend_labels=None):
    """

    :param ax:
    :param pos_y:
    :param neg_y:
    :param data:
    :param hue: if hue is a float-- it's a gradient specifier
        if hue is a 2-tuple-- it's a gradient specifier with different-colored endpoints
        if hue is a tuple of 3-tuples-- treat it AS a color specifier
    :param units:
    :param threshold:
    :param legend_labels: if present, use as an iterable to assign labels to bars for legend purposes
    :return:
    """
    poss = sum([k for k in data if k > 0])
    negs = sum([k for k in data if k < 0])

    total = poss + negs

    left = 0.0
    right = 0.0

    if isinstance(hue, float):
        colors = color_range(len(data), hue)
    elif len(hue) == 2 and isinstance(hue[0], float):
        colors = color_range(len(data), hue)
    else:
        for k in hue:
            assert len(k) == 3, 'invalid color specifier'
        colors = (k for k in hue)

    total_y = None  # where to put the net marker
    total_ha = 'left'  #

    # position the net total indicator
    if poss != 0:
        if negs != 0:
            if total > 0:
                # total indicator is aligned with neg
                total_y = neg_y  # - 0.1
                if total > poss / 2:
                    total_ha = 'right'
            else:
                total_y = pos_y  # + 0.1
                if total > negs / 2:
                    total_ha = 'right'
        else:
            neg_y = pos_y
    else:
        if negs == 0:
            return pos_y  # nothing to do
        neg_y = pos_y

    # print('NY: %f  PY: %f  TY: %f  ha: %s' % (neg_y, pos_y, total_y, total_ha))

    # sparsify labels-- don't pile labels up if there are multiple tiny segments. see _label_segment
    sparse_label_flag_pos = True
    sparse_label_flag_neg = True

    for (i, d) in enumerate(data):
        color = next(colors)
        if legend_labels is None:
            _thelabel = chr(ord('A') + i)
        else:
            _thelabel = ''
        if d > 0:
            patch = ax.barh(pos_y, d, color=color, align='center', left=right, height=1)
            sparse_label_flag_pos = label_segment(patch[0], d, _thelabel, threshold, sparse_label_flag_pos)
            right += d
        elif d < 0:
            left += d
            patch = ax.barh(neg_y, abs(d), color=color, align='center', left=left, height=1)
            sparse_label_flag_neg = label_segment(patch[0], d, _thelabel, threshold, sparse_label_flag_neg)
        else:  # d == 0
            continue
        if legend_labels is not None:
            try:
                patch.set_label(legend_labels[i])
            except IndexError:
                pass

    if units is None:
        unitstring = ''
    else:
        unitstring = ' [%s]' % units

    if poss != 0:
        ax.text(right, pos_y, ' %6.3g%s' % (right, unitstring), fontsize=12, fontweight='bold',
                ha='left', va='center')

    if negs != 0:
        ax.text(right, neg_y, ' %6.3g%s' % (left, unitstring), ha='left', va='center', fontsize=12, fontweight='bold')

    if total_y is not None:
        ax.plot(total, total_y, 'd', markersize=10, markeredgecolor='k', markeredgewidth=0.5)
        ax.barh(total_y, total, color=(0.74, 0.74, 0.74), height=0.65, edgecolor='none')
        ax.text(total, total_y, '   %6.3g   ' % total, ha=total_ha, va='center')

    return neg_y - 0.55


def stack_bar(ax, data, hue, units, title='Contribution Analysis', subtitle='by stage', **kwargs):
    """
    Draws a single stack bar chart on the existing axes.
    ax = axes
    data = ordered data corresponding to the above contributions
    hue = float [0..1] used as the base color in shading the bars
          hue can also be (hue1, hue2), in which case the bars follow a gradient from hue1 to hue2
    units = appended to total in square brackets if present
    title, subtitle- drawn on graph

    If several stack bar charts are to be shown side by side, they should all have the data ordered the same.
    """
    stack_bars(ax, [data], hue, [units], title=title, subtitle=subtitle, **kwargs)


def stack_bars(ax, series, hue, units, labels=None, title='Scenario Analysis', subtitle='by stage',
               center_scale=False, **kwargs):
    """

    :param ax:
    :param series: must be a list of iterables of numerical data, presumably all the same length
    :param hue:
    :param units:
    :param labels: None, or a list of text strings to label each bar set
    :param title:
    :param subtitle:
    :return:
    """
    # first, determine left and right ranges
    left = 0.0
    right = 0.0
    for data in series:
        poss = sum([k for k in data if k > 0])
        negs = sum([k for k in data if k < 0])

        if poss > right:
            right = poss
        if negs < left:
            left = negs

    if center_scale:
        right = max([abs(right), abs(left)])
        if left < 0:
            left = -right

    data_range = right - left
    threshold = 0.07 * data_range
    top = 1.1
    bar_skip = 1.6

    # defaults
    pos_y = .55
    btm = 0

    # make the plots
    for i, data in enumerate(series):
        neg_y = pos_y - 1
        btm = _one_bar(ax, pos_y, neg_y, data, hue, units[i], threshold, **kwargs)
        if labels is not None:
            ax.text(left, pos_y, '%s  ' % labels[i], ha='right', va='center', fontsize=12)
        pos_y = btm - bar_skip

    # common to full plot

    ax.text(0, 2.3, subtitle, color=(0.3, 0.3, 0.3), size='small')
    ax.text(0, 2.8, title, size='large')

    ax.spines['top'].set_visible(False)
    ax.spines['bottom'].set_visible(False)
    ax.spines['left'].set_visible(False)
    ax.spines['right'].set_visible(False)

    ax.set_yticks([])
    if data_range == 0:
        ax.set_xticks([])
        ax.set_xlim(left, left + 1)
        ax.text(0, 1.1, 'No impacts', ha='left', va='bottom')
        ax.set_ylim(1, 3.6)
    else:
        ax.plot((0, 0), (btm, top), 'k-')  # line at origin

        ax.xaxis.set_ticks_position('bottom')
        ax.set_xlim(left - 0.02 * data_range, right + 0.08 * data_range)
        ax.set_ylim(btm, 3.6)


def stack_bar_figure(results, stages, hues=None, color_dict=None, legend=False):
    """
    Create a stack bar figure.
    :param results:
    :param stages:
    :param hues: by default, each stack bar is a gradient of a single hue- supply an array of hues matching length of
     results array, or else hue is generated automatically from quantity uuid / ext ref
    :param color_dict: optionally supply a dict mapping stage name to rgb triple, instead of hue gradient
    :param legend: [False] if true, draw a legend below the bottom bar
    :return:
    """

    if hues is None:
        hues = [None] * len(results)

    units = [r.quantity.unit for r in results]

    # prepare data, determine figure height
    height = 0
    data = [None] * len(results)
    for i, r in enumerate(results):
        data[i] = r.contrib_query(stages)
        height += 1.8
        if has_pos_neg(r):
            height += 0.4

    fig = plt.figure(figsize=(8, height))

    if color_dict is None:
        colors = None
    else:
        colors = [color_dict[s] for s in stages]

    ax = []

    for i, r in enumerate(results):
        ax.append(fig.add_subplot(len(results), 1, i + 1))

        quantity = r.quantity

        if colors is None:
            hue = hues[i]
            if hue is None:
                hue = hue_from_string('%.4s' % quantity.uuid)
        else:
            hue = colors

        stack_bar(ax[i], data[i], hue, units[i], quantity['Name'], quantity['Indicator'], legend_labels=stages)

    def _zero_pos(_a):
        _xl = _a.get_xlim()
        return max(_xl) / (max(_xl) - min(_xl))

    _min = min(_zero_pos(a) for a in ax)

    for a in ax:
        """
        need to adjust axes limits to align zeros
        """
        _xl = a.get_xlim()
        _low = max(_xl) * (_min - 1.0) / _min
        a.set_xlim(_low, max(_xl))

    if legend:
        ax[0].legend(loc='upper right', bbox_to_anchor=(0, 1))

    return fig


def _stackbar_subfig_height(results):
    """
    returns the height of
    :param results: an array of LciaResult objects
    :return:
    """
    height = 0

    for r in results:
        if has_nonzero(r):
            height += 0.9
            if has_pos_neg(r):
                height += 0.7
        else:
            height += 0.6
    return height


def _scenario_fig_height(results, scenario=False):
    height = 0
    if scenario:
        for m, res in enumerate(results):
            height += _stackbar_subfig_height(res)
        height += len(results[0]) - 1
    else:
        for m, res in enumerate(results):
            height += _stackbar_subfig_height([res])
            height += 1

    print('### Height is: %f' % height)
    return height


def scenario_compare_figure(results, stages, hues=None, scenarios=None, savefile=None, center_scale=False,
                            figwidth=8):
    """
    A bit of a swiss army knife.

    if scenarios is None, results is an n-array of LciaResult objects.  Creates an n-subplot figure of
    stacked pos/neg bars, annotated by reference to k ordered stages.

     if scenarios is not none, it must be an m-array of scenario names or labels.  results is interpreted as
     an m-array of n-arrays of LciaResult objects. Creates an n-subplot figure of m-stacked pos/neg bars, each
     featuring k stages

     this needs to be done vertical, with the stages to the side

     or else the waterfall charts need to be scaled
     waterfall wasn't made that way

    :param results: either a n-array or a m-by-n array of results
    :param stages: k query terms to results
    :param hues: for n quantities represented (quick-compute from UUID)
    :param scenarios: None or m-array
    :param savefile: if present, save the fig to the specified file
    :return:
    """

    if scenarios is None:
        height = _scenario_fig_height(results, scenario=False)
        fig = plt.figure(figsize=(figwidth, height))
        quantities = [r.quantity for r in results]

    else:
        height = _scenario_fig_height(results, scenario=True)
        fig = plt.figure(figsize=(figwidth, height))
        quantities = [r.quantity for r in results[0]]

    if hues is None:
        hues = [None] * len(quantities)

    for n, quantity in enumerate(quantities):
        hue = hues[n]
        ax = fig.add_subplot(len(quantities), 1, n + 1)

        if hue is None:
            hue = hue_from_string('%.4s' % quantity.uuid)

        if scenarios is None:
            series = results[n].contrib_query(stages)
            stack_bars(ax, [series], hue, [quantity.unit],
                       title=quantity['Name'], subtitle=quantity['Indicator'],
                       center_scale=center_scale)

        else:
            series = [r[n].contrib_query(stages) for r in results]
            units = [r[n].quantity.unit for r in results]

            stack_bars(ax, series, hue, units, labels=scenarios,
                       title=quantity['Name'], subtitle=quantity['Indicator'],
                       center_scale=center_scale)

        """
        if n == 0:
            colors = [k for k in color_range(len(stages), hue)]
            handles = [mpatches.Patch(color=colors[i], label='%s -- %s' % (chr(ord('A') + i), k))
                       for i, k in enumerate(stages)]

            ax.legend(handles=handles, bbox_to_anchor=(1.75, 0.75))
            """
    leg = 'Stages: \n' + '\n'.join(['%s -- %s' % (chr(ord('A') + i), k) for i, k in enumerate(stages)])
    print(leg)

    if savefile is not None:
        save_plot(savefile, close_after=False)

    return fig



def transport_model(modeled_flow):
    return 'Transport Mix - %s' % modeled_flow.name


def activity_model_ref(modeled_fragment):
    n = 'Activity model - %s (%s)' % (modeled_fragment['Name'], modeled_fragment.get('scope'))
    n = n.replace('/', '_')
    return n


def logistics_summary_ref(sc_frag):
    return '%s Logistics' % sc_frag.name



from __future__ import division
import colorsys
import matplotlib as mpl
import matplotlib.pyplot as plt
from textwrap import wrap

import os

# grumble something about breaking styles for no good reason with 2.0.0
mpl.rcParams['patch.force_edgecolor'] = True
mpl.rcParams['errorbar.capsize'] = 3
mpl.rcParams['grid.color'] = 'k'
mpl.rcParams['grid.linestyle'] = ':'
mpl.rcParams['grid.linewidth'] = 0.5
mpl.rcParams['lines.linewidth'] = 1.0
mpl.rcParams['axes.autolimit_mode'] = 'round_numbers'
mpl.rcParams['axes.xmargin'] = 0
mpl.rcParams['axes.ymargin'] = 0


prefab_colors = ((0.1, 0.6, 0.7), (0.9, 0.3, 0.4), (0.3, 0.9, 0.6), (0.1, 0.1, 0.8), (0.4, 0.9, 0.3), (0.3, 0.4, 0.9))
net_color = (0.6, 0.6, 0.6)


def linspace(a, b, num):
    if num < 2:
        yield a
    else:
        dif = float(b - a) / (num - 1)
        for i in range(num - 1):
            yield a
            a += dif
    yield b


def color_range(num, hue, sat=0.5):
    """
    Generator for a series of num RGB tuples, derived from the color provided in hue.
    If hue is a 2-tuple, provides a gradation from [0] to [1]. Uses either hsv or hsl to rgb.
    """
    if not isinstance(hue, tuple):
        hue = (hue, hue)
    h = linspace(hue[0], hue[1], num)
    v = linspace(0.95, 0.45, num)
    for i in range(num):
        yield colorsys.hsv_to_rgb(next(h), sat, next(v))


def hue_from_string(s, base=16):
    hue = int(s, base)
    return hue / (base**len(s))


def _label_bar(patch, value=None, label=None, valueformat='%4.3g', labelformat='%s'):
    bl = patch.get_xy()
    x = 0.5 * patch.get_width() + bl[0]
    y = 0.5 * patch.get_height() + bl[1]
    if value is not None:
        patch.axes.text(x, y, valueformat % value, ha='center', va='center')
    if label is not None:
        patch.axes.text(x, y + (0.52 * patch.get_height()), labelformat % label, ha='center', va='bottom')


def label_segment(patch, data, label, threshold, sparse_flag):
    """
    Label the segment if it's big enough. The threshold is the minimum size for a value statement; one-quarter the
    threshold is the minimum size for a bar label. Segments smaller than one-quarter the threshold are considered
    non-sparse, and if there is more than one non-sparse segment in succession, only the first will be labeled.
    A single non-sparse segment resets the labeling.  Positive and negative sparse_flags should be tracked separately.
    :param patch:
    :param data:
    :param label:
    :param threshold:
    :param sparse_flag:
    :return:
    """
    if abs(data) > threshold:
        _label_bar(patch, value=data, label=label)
    if sparse_flag or abs(data) > (threshold / 4):
        _label_bar(patch, label=label)
    if abs(data) < (threshold / 4):
        return False
    return True


def label_vbar(patch, value, valueformat='%6.3g', sep=0):
    (y0, y1) = patch.axes.get_ylim()
    vgap = 0.02 * (y1 - y0)
    bl = patch.get_xy()
    x = 0.5 * patch.get_width() + bl[0]

    y = bl[1] + patch.get_height() + sep + vgap

    patch.axes.text(x, y, valueformat % value, ha='center', va='bottom')


def has_nonzero(res):
    data = sum([abs(i) for i in res.contrib_query()])
    return data != 0


def has_pos_neg(res):
    """
    Returns true if the input contains both positive and negative data points
    :param res:
    :return:
    """
    data = res.contrib_query()
    try:
        poss = next(k for k in data if k > 0)
        negs = next(k for k in data if k < 0)
    except StopIteration:
        return False
    if poss != 0 and negs != 0:
        return True
    return False


def save_plot(file, close_after=True, transparent=True, format=None):

    kwargs = {
        'transparent': transparent,
        'format': format,
        'bbox_inches': 'tight'
    }

    if format is None:
        n, f = os.path.splitext(file)
        if f:
            format = f[1:]
            if format == 'png':
                kwargs['dpi'] = 300
        else:
            format = 'eps'

    plt.savefig(file, **kwargs)
    if close_after:
        plt.close()


def standard_labels(ax, stages, ticks=None, width=25, rotate=True):
    """

    :param ax: axes to modify
    :param stages: tick labels
    :param ticks: locations of ticks [default: range(len(stages))]
    :param width: [25] text wrap width
    :param rotate: [True] whether to rotate long labels.  specify bool or rotation argument.
    :return:
    """
    rotation = 0
    if ticks is None:
        ticks = range(len(stages))
    ax.set_xticks(ticks)
    if max([len(l) for l in stages]) > 12:
        labels = ['\n'.join(wrap(l, width)) for l in stages]
        if rotate:
            if isinstance(rotate, bool):
                rotation = 90
            else:
                rotation = rotate
    else:
        labels = stages

    ax.set_xticklabels(labels, rotation=rotation)


def open_ylims(ax, margin=0.1):
    """
    workaround for buggy Axes.margins() (https://github.com/matplotlib/matplotlib/pull/7995 and others)
    :param ax:
    :param margin:
    :return:
    """
    bottom, top = ax.get_ylim()
    yr = top - bottom
    if bottom < 0 < top:
        bottom -= margin * yr
        top += margin * yr
        ax.set_ylim(bottom=bottom, top=top)


"""
people = ('A', 'B', 'C', 'D', 'E', 'F', 'G', 'H')
segments = 4

# generate some multi-dimensional data & arbitrary labels
data = 3 + 10 * np.random.rand(segments, len(people))
percentages = (np.random.randint(5,20, (len(people), segments)))
y_pos = np.arange(len(people))

fig = plt.figure(figsize=(10, 8))
ax = fig.add_subplot(111)

colors = 'rgbwmc'
patch_handles = []
left = np.zeros(len(people))  # left alignment of data starts at zero
for i, d in enumerate(data):
    patch_handles.append(ax.barh(y_pos, d,
                                 color=colors[i % len(colors)], align='center',
                                 left=left))
    # accumulate the left-hand offsets
    left += d

# go through all of the bar segments and annotate
for j in range(len(patch_handles)):
    for i, patch in enumerate(patch_handles[j].get_children()):
        bl = patch.get_xy()
        x = 0.5*patch.get_width() + bl[0]
        y = 0.5*patch.get_height() + bl[1]
        ax.text(x, y, "%d%%" % (percentages[i, j]), ha='center')

ax.set_yticks(y_pos)
ax.set_yticklabels(people)
ax.set_xlabel('Distance')

plt.show()
"""



"""
Functions for creating tables for useful / important comparisons.  These are analogous to charts in that they
are forms of output and it's not clear where they belong.

Lists of tabular outputs:
 * process or fragment Inventory

 * compare process inventories
 * compare allocations of a multioutput process

 * compare LCIA factors for different methods
 * compare an LCIA method with the components of one or more Lcia Results using it

Here's another thing: right now I'm using dynamic grid to show these in the window... but wouldn't it perhaps be
preferable to use pandas? doesn't pandas afford all sorts of useful features, like ...
um...
what is pandas good for again? for working with data frames. Not necessarily for creating data frames.

Most likely, I could modify dynamic_grid to *return* a dataframe instead of drawing a table.
"""

from collections import defaultdict
from pandas import DataFrame


def printable(tup, width=8):
    out = []
    for k in tup:
        if isinstance(k, str):
            out.append(k)
        elif k is None:
            out.append('')
        else:
            try:
                g = '%*.3g' % (width, k)
            except TypeError:
                g = '%*.*s' % (width, width, '----')
            out.append(g)
    return tuple(out)


class BaseTableOutput(object):
    """
    A prototype class for storing and returning tabular information.  This should ultimately be adopted in places
    where dynamic_grids are used, or where TeX or excel tables are produced (like in lca_matrix foreground output
    generators) but for now it is just being used to provide some separation of concerns for the flowables super-grid.

    At the heart is a dict whose key is a 2-tuple of (row signifier, column index).  The row signifier can be any
    hashable object, but the column indices are always sequential. re-ordering columns is something we do not feel
    particularly like enabling at the present time.

    The user creates the table with initialization parameters as desired, and then builds out the table by adding
    columns in sequence.

    The table has an inclusion criterion for the iterables (which could be None)-- if the criterion is met, the object
    is added; if not, it is skipped.  The criterion can change, but (since the table contents are static) this will not
    result in columns being re-iterated.

    Subclasses MAY redefine:
     _returns_sets: determines whether each grid item is singly or multiply valued

    Subclasses MUST implement:
     _near_headings -- column names for left-side headings
     _generate_items(col) -- argument is a column iterable - generates items
     _pull_row_from_item(item) -- argument is one of the objects returned by the column iteration, returns row key
     _extract_data_from_item -- argument is an dict from the grid dict, returns either a dict or an immutable object

    """

    _near_headings = '',  # should be overridden
    _far_headings = '', # should be overridden
    _returns_sets = False

    def _pull_row_from_item(self, item):
        """
        Returns the row tuple from an item, for insertion into the rows set. meant to be overridden
        :param item:
        :return: always a tuple.  default item,
        """
        row = item
        # if not self._returns_sets:
        return row,

    def _pull_note_from_item(self, item):
        """
        Returns the "long" / descriptive text appended to the right-hand side of the table. should return a str.
        Only used if _returns_sets is false (otherwise, the sets indicate the row + subrow labels)
        This is may turn out to be totally silly / pointless.
        :param item:
        :return:
        """
        return ''

    def _generate_items(self, iterable):
        """
        yields the items from a column entry. Meant to be overridden.
        :param iterable:
        :return:
        """
        for item in iterable:
            if self._criterion(item):
                yield item

    def _extract_data_from_item(self, item):
        """
        note: dict item is a list of components
        Determines how to get the data point from the item/list. Meant to be overridden.
        If self._returns_sets is true, should return a dict. Else should return an immutable.
        :param item:
        :return: a string
        """
        return item

    def _header_row(self):
        """
        Returns a tuple of columns for the header row
        :return:
        """
        header = self._near_headings
        for i, _ in enumerate(self._columns):
            header += ('C%d' % i),
        header += self._far_headings  # placeholder for row notes / subitem keys
        return header

    def _build_near_header(self, row, prev):
        the_row = []
        for i, _ in enumerate(self._near_headings):
            if prev is not None:
                if prev[i] == row[i]:
                    the_row.append('""')
                    continue
            the_row.append('%s' % row[i])
        return the_row

    def _build_row(self, row, prev=None):
        """
        Returns a single row as a tuple.
        :param row:
        :param prev: [None] previous row printed (input, not output). Used to suppress header output for repeat entries.
        :return:
        """
        # first build the near header
        the_row = self._build_near_header(row, prev)

        data_keys = set()
        data_vals = []

        # first pass: get all the data / keys
        for i, _ in enumerate(self._columns):
            data = self._extract_data_from_item(self._d[row, i])
            if isinstance(data, dict):
                if not self._returns_sets:
                    raise TypeError('multiple values returned but subclass does not allow them!')
                for k in data.keys():
                    data_keys.add(k)
            data_vals.append(data)

        # second pass: build the sub-table by rows
        if self._returns_sets:
            the_rows = []
            _ftt = True  # first time through
            keys = tuple(sorted(data_keys, key=lambda x: x[-2]))
            for k in keys:
                if not _ftt:
                    the_row = ['' for i in range(len(self._near_headings))]
                for i, _ in enumerate(self._columns):
                    if k in data_vals[i]:
                        the_row.append(data_vals[i][k])
                    else:
                        the_row.append(None)
                the_row.append(k)
                if _ftt:
                    the_row.append(self._notes[row])
                else:
                    the_row.append('')
                the_rows.append(the_row)
                _ftt = False
            return the_rows
        else:
            the_row.extend(data_vals)
            # add notes
            the_row.append(self._notes[row])
            return the_row

    def __init__(self, *args, criterion=None):
        """
        Provide 0 or more positional arguments as data columns; add data columns later with add_column(arg)

        :param args: sequential data columns
        :param criterion: A callable expression that returns true if a given
        """
        self._d = defaultdict(list)

        if callable(criterion):
            self._criterion = criterion
        else:
            if criterion is not None:
                print('Ignoring non-callable criterion')

            self._criterion = lambda x: True

        self._rows = set()  # set of valid keys to dict
        self._notes = dict()
        self._columns = []  # list of columns in the order added
        # a valid reference consists of (x, y) where x in self._rows and y < len(self._columns)

        for arg in args:
            self.add_column(arg)

    def _add_rowitem(self, col_idx, item, row=None):
        if row is None:
            row = self._pull_row_from_item(item)
        self._rows.add(row)
        if row not in self._notes:
            self._notes[row] = self._pull_note_from_item(item)
        self._d[row, col_idx].append(item)

    def add_column(self, arg):
        col_idx = len(self._columns)
        for k in self._generate_items(arg):
            self._add_rowitem(col_idx, k)
        self._columns.append(arg)

    def _sorted_rows(self):
        for row in sorted(self._rows, key=lambda x: tuple([str(k) for k in x])):
            yield row

    def text(self, width=10, hdr_width=24, max_width=112, expanded=True):
        """
        Outputs the table in text format
        :return: nothing.
        """
        header = self._header_row()
        prev = None
        body = []
        width = max(6, width)

        # widths of each column
        wds = [len(header[i]) for i in range(len(self._near_headings))]

        # determine column widths
        for row in self._sorted_rows():
            prt_row = self._build_row(row, prev=prev)

            if self._returns_sets:
                wds = [min(max(wds[i], len('%s' % prt_row[0][i])), hdr_width) for i in range(len(self._near_headings))]
            else:
                wds = [min(max(wds[i], len('%s' % prt_row[i])), hdr_width) for i in range(len(self._near_headings))]

            body.append(prt_row)
            prev = row

        # build display string
        rem_width = max_width
        fmt = ''
        for i in wds:
            rem_width -= i
            fmt += '%%-%d.%ds ' % (i, i)
            rem_width -= 1

        for i in range(len(self._columns)):
            rem_width -= width
            fmt += '%%-%d.%ds ' % (width, width)
            rem_width -= 1

        if rem_width < 0:
            # uh oh negative rem width: widen freely; set remainder to 10 chars
            max_width -= (rem_width - 10)
            rem_width = 10

        fmt += '%%-%d.%ds' % (rem_width, rem_width)

        if self._returns_sets:
            fmt += ' %s'

        print(fmt % header)
        print('-' * max_width)

        for row in body:
            if self._returns_sets:
                for subrow in row:  # sorted(row, key=lambda x: x[-2])
                    print(fmt % printable(subrow, width=width))
            else:
                print(fmt % printable(row, width=width))

        print(fmt % header)
        print('\nColumns:')
        for i, c in enumerate(self._columns):
            print('C%d: %s' % (i, c))

    def dataframe(self):
        df = DataFrame(columns=self._header_row())
        prev = None
        for row in self._sorted_rows():
            if self._returns_sets:
                for r in self._build_row(row):
                    d = dict(zip(self._header_row(), printable(r)))
                    df = df.append(d, ignore_index=True)
            else:
                d = dict(zip(self._header_row(), printable(self._build_row(row, prev=prev))))
                df = df.append(d, ignore_index=True)
            prev = row

        return df

    def to_excel(self, xl_writer, sheetname, width_scaling=0.75):
        """
        Must supply a pandas XlsxWriter. This routine does not save the document.
        :param xl_writer:
        :param sheetname:
        :param width_scaling:
        :return:
        """
        df = self.dataframe()
        df.to_excel(xl_writer, sheet_name=sheetname)
        sht = xl_writer.sheets[sheetname]

        for k in self._near_headings + self._far_headings:
            ix = df.columns.tolist().index(k) + 1
            mx = max([7, width_scaling * df[k].astype(str).str.len().max()])
            sht.set_column(ix, ix, width=mx)
